# Copyright (c) 2022 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import getpass
import io
import os
import socket
import shutil
import tempfile
import typing
import uuid

import testtools

import tobiko
from tobiko.openstack import stacks
from tobiko.openstack import topology
from tobiko.shell import sh
from tobiko.shell import ssh


class LocalTempDirFixture(tobiko.SharedFixture):

    path: typing.Optional[str] = None

    def setup_fixture(self):
        self.path = self.create_dir()

    def cleanup_fixture(self):
        path = self.path
        if path is not None:
            try:
                self.delete_dir()
            finally:
                del self.path

    def create_dir(self) -> str:
        return tempfile.mkdtemp()

    def delete_dir(self):
        if os.path.isdir(self.path):
            shutil.rmtree(self.path)


class LocalShellConnectionTest(testtools.TestCase):

    @property
    def ssh_client(self) -> ssh.SSHClientType:
        return False

    @property
    def connection(self) -> sh.ShellConnection:
        return sh.shell_connection(self.ssh_client)

    connection_class = sh.LocalShellConnection

    def test_shell_connection(self):
        connection = sh.shell_connection(self.ssh_client)
        self.assertIsInstance(connection, self.connection_class)
        self.assertIs(self.connection, connection)

    @property
    def is_local(self) -> bool:
        return True

    def test_is_local(self):
        self.assertIs(self.is_local, self.connection.is_local)

    def test_is_local_connection(self):
        is_local = sh.is_local_connection(self.ssh_client)
        self.assertIs(self.is_local, is_local)

    @property
    def hostname(self) -> str:
        return socket.gethostname()

    def test_hostname(self):
        self.assertEqual(self.hostname, self.connection.hostname)

    def test_connection_hostname(self):
        hostname = sh.connection_hostname(self.ssh_client)
        self.assertEqual(self.hostname, hostname)

    @property
    def login(self) -> str:
        return f"{self.username}@{self.hostname}"

    def test_login(self):
        self.assertEqual(self.login, self.connection.login)

    def test_connection_login(self):
        login = sh.connection_login(self.ssh_client)
        self.assertEqual(self.login, login)

    @property
    def username(self) -> str:
        return getpass.getuser()

    def test_username(self):
        self.assertEqual(self.username, self.connection.username)

    def test_connection_username(self):
        username = sh.connection_username(self.ssh_client)
        self.assertEqual(self.username, username)

    @property
    def is_cirros(self) -> bool:
        return False

    def test_is_cirros(self):
        self.assertIs(self.is_cirros, self.connection.is_cirros)

    def test_is_cirros_connection(self):
        is_cirros = sh.is_cirros_connection(self.ssh_client)
        self.assertIs(self.is_cirros, is_cirros)

    @property
    def local_connection(self) -> sh.LocalShellConnection:
        return sh.local_shell_connection()

    def test_get_file(self):
        local_file = os.path.join(self.local_connection.make_temp_dir(),
                                  'local_file')
        remote_file = os.path.join(self.connection.make_temp_dir(),
                                   'remote_file')
        text = str(uuid.uuid4())
        sh.execute(f"echo '{text}' > '{remote_file}'",
                   ssh_client=self.ssh_client)
        self.assertFalse(os.path.isfile(local_file))
        sh.get_file(local_file=local_file,
                    remote_file=remote_file,
                    connection=self.ssh_client)
        self.assertTrue(os.path.isfile(local_file), 'file not copied')
        with io.open(local_file, 'rt') as fd:
            self.assertEqual(f'{text}\n', fd.read())

    def test_put_file(self):
        local_file = os.path.join(self.local_connection.make_temp_dir(),
                                  'local_file')
        remote_file = os.path.join(self.connection.make_temp_dir(),
                                   'remote_file')
        text = str(uuid.uuid4())
        with io.open(local_file, 'wt') as fd:
            fd.write(text)
        self.assertRaises(sh.ShellCommandFailed,
                          sh.execute, f"cat '{remote_file}'",
                          ssh_client=self.ssh_client)
        sh.put_file(remote_file=remote_file,
                    local_file=local_file,
                    connection=self.ssh_client)
        output = sh.execute(f"cat '{remote_file}'",
                            ssh_client=self.ssh_client).stdout
        self.assertEqual(text, output)

    def test_put_files(self):
        local_dir = self.local_connection.make_temp_dir()
        local_file = os.path.join(local_dir, 'some_file')
        remote_dir = self.connection.make_temp_dir()
        remote_file = os.path.join(remote_dir,
                                   os.path.basename(local_dir),
                                   'some_file')
        text = str(uuid.uuid4())
        with io.open(local_file, 'wt') as fd:
            fd.write(text)
        self.assertRaises(sh.ShellCommandFailed,
                          sh.execute, f"cat '{remote_file}'",
                          ssh_client=self.ssh_client)
        sh.put_files(local_dir,
                     remote_dir=remote_dir,
                     connection=self.ssh_client)
        output = sh.execute(f"cat '{remote_file}'",
                            ssh_client=self.ssh_client).stdout
        self.assertEqual(text, output)

    def test_open_file(self):
        temp_file = self.connection.make_temp_file()
        with self.connection.open_file(temp_file, 'wt') as fd:
            fd.write('something')
        with self.connection.open_file(temp_file, 'rt') as fd:
            self.assertIn(fd.read(), [b'something',
                                      'something'])

    def test_make_temp_file(self):
        temp_file = self.connection.make_temp_file()
        self.assertTrue(self.connection.exists(temp_file))
        self.assertTrue(self.connection.is_file(temp_file))
        self.assertFalse(self.connection.is_directory(temp_file))

    def test_make_temp_dir(self):
        temp_file = self.connection.make_temp_dir()
        self.assertTrue(self.connection.exists(temp_file))
        self.assertFalse(self.connection.is_file(temp_file))
        self.assertTrue(self.connection.is_directory(temp_file))


class SSHShellConnectionTest(LocalShellConnectionTest):
    connection_class = sh.SSHShellConnection
    server = tobiko.required_fixture(stacks.UbuntuMinimalServerStackFixture)

    @property
    def ssh_client(self) -> ssh.SSHClientFixture:
        ssh_client = ssh.ssh_proxy_client()
        if isinstance(ssh_client, ssh.SSHClientFixture):
            return ssh_client

        nodes = topology.list_openstack_nodes()
        for node in nodes:
            if isinstance(node.ssh_client, ssh.SSHClientFixture):
                return node.ssh_client

        return self.server.ssh_client

    @property
    def is_local(self) -> bool:
        return False

    @property
    def hostname(self) -> str:
        return sh.get_hostname(ssh_client=self.ssh_client)

    @property
    def username(self) -> str:
        return self.ssh_client.setup_connect_parameters()['username']


class CirrosShellConnectionTest(SSHShellConnectionTest):
    connection_class = stacks.CirrosShellConnection
    server = tobiko.required_fixture(stacks.CirrosServerStackFixture)

    @property
    def ssh_client(self) -> ssh.SSHClientFixture:
        return self.server.ssh_client

    @property
    def is_cirros(self) -> bool:
        return True

    @tobiko.skip('not implemented')
    def test_open_file(self):
        pass



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os.path

import tobiko
from tobiko.shell import sh
from tobiko import tripleo


class ShiftStackCloudsFileFixture(tobiko.SharedFixture):

    def __init__(self,
                 local_clouds_file_path: str = None,
                 remote_clouds_file_path: str = None,
                 connection: sh.ShellConnection = None):
        super(ShiftStackCloudsFileFixture, self).__init__()
        self._local_clouds_file_path = local_clouds_file_path
        self._remote_clouds_file_path = remote_clouds_file_path
        self._connection = connection

    @property
    def local_clouds_file_path(self) -> str:
        if self._local_clouds_file_path is None:
            self._local_clouds_file_path = tobiko.tobiko_config_path(
                tobiko.tobiko_config().shiftstack.local_clouds_file_path)
        assert isinstance(self._local_clouds_file_path, str)
        return self._local_clouds_file_path

    @property
    def remote_clouds_file_path(self) -> str:
        if self._remote_clouds_file_path is None:
            self._remote_clouds_file_path = (
                tobiko.tobiko_config().shiftstack.remote_clouds_file_path)
        assert isinstance(self._remote_clouds_file_path, str)
        return self._remote_clouds_file_path

    @property
    def connection(self) -> sh.ShellConnection:
        if self._connection is None:
            self._connection = sh.shell_connection(
                tripleo.undercloud_ssh_client())
        assert isinstance(self._connection, sh.ShellConnection)
        return self._connection

    def setup_fixture(self):
        tobiko.makedirs(os.path.dirname(self.local_clouds_file_path),
                        exist_ok=True)
        self.connection.get_file(
            remote_file=self.remote_clouds_file_path,
            local_file=self.local_clouds_file_path)


def get_clouds_file_path() -> str:
    return tobiko.setup_fixture(
        ShiftStackCloudsFileFixture).local_clouds_file_path



# Copyright (c) 2021 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from octaviaclient.api import exceptions

import tobiko


class RequestException(tobiko.TobikoException):
    message = ("Error while sending request to server "
               "(command was '{command}'): {error}")


class TimeoutException(tobiko.TobikoException):
    message = "Timeout exception: {reason}"


OctaviaClientException = exceptions.OctaviaClientException


class RoundRobinException(tobiko.TobikoException):
    message = "Round robin exception: {reason}"


class TrafficTimeoutError(tobiko.TobikoException):
    message = "Traffic timeout error: {reason}"


class AmphoraMgmtPortNotFound(tobiko.TobikoException):
    message = "Amphora's network management port was not found: {reason}"



# Copyright (c) 2020 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os
import typing

from oslo_log import log

import tobiko
from tobiko.shell.ssh import _client


LOG = log.getLogger(__name__)


DEFAULT_SSH_KEY_FILE = "~/.ssh/id_rsa"


def get_key_file(ssh_client: _client.SSHClientFixture,
                 key_file: str = DEFAULT_SSH_KEY_FILE):
    return tobiko.setup_fixture(
        GetSSHKeyFileFixture(ssh_client=ssh_client,
                             remote_key_file=key_file)).key_file


class GetSSHKeyFileFixture(tobiko.SharedFixture):

    key_file = None

    def __init__(self, ssh_client: _client.SSHClientFixture,
                 remote_key_file: str = DEFAULT_SSH_KEY_FILE):
        super(GetSSHKeyFileFixture, self).__init__()
        self.ssh_client = ssh_client
        self.remote_key_file = remote_key_file

    def setup_fixture(self):
        client = self.ssh_client.connect()
        _, stdout, stderr = client.exec_command('hostname')
        remote_hostname = stdout.read().strip().decode()
        if not remote_hostname:
            error = stderr.read()
            raise RuntimeError(
                "Unable to get hostname from proxy jump server:\n"
                f"{error}")

        _, stdout, stderr = client.exec_command(
            f"cat {self.remote_key_file}")
        private_key = stdout.read()
        if not private_key:
            error = stderr.read()
            LOG.error("Unable to get SSH private key from proxy jump "
                      f"server:\n{error}")
            return

        _, stdout, stderr = client.exec_command(
            f"cat {self.remote_key_file}.pub")
        public_key = stdout.read()
        if not public_key:
            error = stderr.read()
            LOG.error("Unable to get SSH public key from proxy jump "
                      f"server:\n{error}")
            return

        key_file = tobiko.tobiko_config_path(
            f"~/.ssh/{os.path.basename(self.remote_key_file)}-" +
            remote_hostname)
        with tobiko.open_output_file(key_file) as fd:
            fd.write(private_key.decode())
        with tobiko.open_output_file(key_file + '.pub') as fd:
            fd.write(public_key.decode())
        self.key_file = key_file


def list_key_filenames() -> typing.List[str]:
    return tobiko.setup_fixture(KeyFilenamesFixture).key_filenames


def list_proxy_jump_key_filenames() -> typing.List[str]:
    return tobiko.setup_fixture(ProxyJumpKeyFilenamesFixture).key_filenames


class KeyFilenamesFixture(tobiko.SharedFixture):

    def __init__(self):
        super().__init__()
        self.key_filenames: typing.List[str] = []

    def setup_fixture(self):
        conf = tobiko.tobiko_config()
        for key_file in tobiko.select_uniques(conf.ssh.key_file):
            if key_file:
                key_file = tobiko.tobiko_config_path(key_file)
                if (key_file not in self.key_filenames and
                        os.path.isfile(key_file)):
                    LOG.info(f"Use local keyfile: {key_file}")
                    self.key_filenames.append(key_file)


class ProxyJumpKeyFilenamesFixture(KeyFilenamesFixture):

    def setup_fixture(self):
        ssh_proxy_client = _client.ssh_proxy_client()
        if ssh_proxy_client is None:
            return

        conf = tobiko.tobiko_config()
        remote_key_files = tobiko.select_uniques(conf.ssh.key_file)
        for remote_key_file in remote_key_files:
            key_file = get_key_file(ssh_client=ssh_proxy_client,
                                    key_file=remote_key_file)
            if (key_file and
                    key_file not in self.key_filenames and
                    os.path.isfile(key_file)):
                LOG.info(f"Use SSH proxy server keyfile: {key_file}")
                self.key_filenames.append(key_file)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import


from oslo_log import log

import tobiko
from tobiko import config
from tobiko.openstack import heat
from tobiko.openstack import octavia
from tobiko.openstack import neutron
from tobiko.openstack.stacks import _hot
from tobiko.openstack.stacks import _neutron
from tobiko.openstack.stacks import _ubuntu
from tobiko.shell import sh

CONF = config.CONF
LOG = log.getLogger(__name__)


class AmphoraIPv4LoadBalancerStack(heat.HeatStackFixture):
    template = _hot.heat_template_file('octavia/load_balancer.yaml')

    vip_network = tobiko.required_fixture(_neutron.NetworkStackFixture)

    #: Floating IP network where the Neutron floating IP are created
    @property
    def floating_network(self) -> str:
        return self.vip_network.floating_network

    @property
    def has_floating_ip(self) -> bool:
        return bool(self.floating_network)

    ip_version = 4

    provider = 'amphora'

    @property
    def vip_subnet_id(self):
        if self.ip_version == 4:
            return self.vip_network.ipv4_subnet_id
        else:
            return self.vip_network.ipv6_subnet_id

    def wait_for_active_loadbalancer(self,
                                     timeout: tobiko.Seconds = None):
        octavia.wait_for_status(status_key=octavia.PROVISIONING_STATUS,
                                status=octavia.ACTIVE,
                                get_client=octavia.get_load_balancer,
                                object_id=self.loadbalancer_id,
                                timeout=timeout)

    def wait_for_update_loadbalancer(self,
                                     timeout: tobiko.Seconds = None):
        octavia.wait_for_status(status_key=octavia.PROVISIONING_STATUS,
                                status=octavia.PENDING_UPDATE,
                                get_client=octavia.get_load_balancer,
                                object_id=self.loadbalancer_id,
                                timeout=timeout)

    def wait_for_octavia_service(self,
                                 interval: tobiko.Seconds = None,
                                 timeout: tobiko.Seconds = None,
                                 client=None):
        for attempt in tobiko.retry(timeout=timeout,
                                    interval=interval,
                                    default_timeout=180.,
                                    default_interval=5.):
            try:
                octavia.list_amphorae(loadbalancer_id=self.loadbalancer_id,
                                      client=client)
            except octavia.OctaviaClientException as ex:
                LOG.debug(f"Error listing amphorae: {ex}")
                if attempt.is_last:
                    raise
                LOG.info('Waiting for the LB to become functional again...')
            else:
                LOG.info('Octavia service is available!')
                break


class AmphoraIPv6LoadBalancerStack(AmphoraIPv4LoadBalancerStack):
    ip_version = 6


class OctaviaOtherServerStackFixture(_ubuntu.UbuntuServerStackFixture):
    pass


class HttpRoundRobinAmphoraIpv4Listener(heat.HeatStackFixture):
    template = _hot.heat_template_file('octavia/listener.yaml')

    loadbalancer = tobiko.required_fixture(
        AmphoraIPv4LoadBalancerStack)

    lb_port = 80

    lb_protocol = 'HTTP'

    @property
    def loadbalancer_id(self):
        return self.loadbalancer.loadbalancer_id

    @property
    def loadbalancer_provider(self):
        return self.loadbalancer.provider

    # Pool attributes
    pool_protocol = 'HTTP'

    lb_algorithm = 'ROUND_ROBIN'

    # healthmonitor attributes
    hm_type = 'HTTP'

    hm_delay = 3

    hm_max_retries = 4

    hm_timeout = 3

    #: whenever to create the health monitor
    has_monitor = True

    @property
    def listener_id(self):
        return self.listener.listener_id

    def wait_for_active_members(self):
        """Wait for all pool members to be active"""

        for member in octavia.list_members(pool_id=self.pool_id):
            self.wait_for_active_member(pool_id=self.pool_id,
                                        member_id=member['id'])

    def wait_for_active_member(self, pool_id, member_id, **kwargs):
        """Wait for the member to be active

        Waits for the member to have an ACTIVE provisioning status.

        :param member_id: the member id.
        :param pool_id: the pool id.
        """
        octavia.wait_for_status(status_key=octavia.PROVISIONING_STATUS,
                                status=octavia.ACTIVE,
                                get_client=octavia.get_member,
                                object_id=pool_id,
                                member_id=member_id, **kwargs)

    def wait_for_members_to_be_reachable(self,
                                         interval: tobiko.Seconds = None,
                                         timeout: tobiko.Seconds = None):

        members = [self.server_stack, self.other_server_stack]

        if len(members) < 1:
            return

        # Wait for members to be reachable from localhost
        last_reached_id = 0
        for attempt in tobiko.retry(
                timeout=timeout,
                interval=interval,
                default_interval=5.,
                default_timeout=members[0].wait_timeout):
            try:
                for member in members[last_reached_id:]:
                    octavia.check_members_balanced(
                        members_count=1,
                        ip_address=member.ip_address,
                        protocol=self.lb_protocol,
                        port=self.lb_port,
                        requests_count=1)
                    last_reached_id += 1  # prevent retrying same member again
            except sh.ShellCommandFailed:
                if attempt.is_last:
                    raise
                LOG.info(
                    "Waiting for members to have HTTP service available...")
                continue
            else:
                break
        else:
            raise RuntimeError("Members couldn't be reached!")

    # Members attributes
    server_stack = tobiko.required_fixture(_ubuntu.UbuntuServerStackFixture)

    other_server_stack = tobiko.required_fixture(
        OctaviaOtherServerStackFixture)

    application_port = 80

    ip_version = 4

    @property
    def pool_id(self):
        return self.pool.pool_id

    @property
    def subnet_id(self):
        network_stack = self.server_stack.network_stack
        if self.ip_version == 4:
            return network_stack.ipv4_subnet_id
        else:
            return network_stack.ipv6_subnet_id

    @property
    def member_address(self) -> str:
        return self.get_member_address(self.server_stack)

    @property
    def other_member_address(self) -> str:
        return self.get_member_address(self.other_server_stack)

    def get_member_address(self, server_stack):
        return str(server_stack.find_fixed_ip(ip_version=self.ip_version))

    @property
    def amphora(self) -> octavia.AmphoraType:
        return octavia.get_master_amphora(
            amphorae=octavia.list_amphorae(
                self.loadbalancer.loadbalancer_id),
            ip_address=self.loadbalancer.floating_ip_address,
            port=self.lb_port,
            protocol=self.lb_protocol)

    @property
    def amphora_mgmt_port(self) -> neutron.PortType:
        """ Get amphora's management network's port

        Attaching a floating ip to the amphora's management port will allow
        SSHing the amphora.
        Disabling the amphora's management port will cause a failover.
        """
        # Calling self.amphora in a loop would decrease performance
        amphora = self.amphora

        for port in neutron.list_ports():
            for fixed_ip in port['fixed_ips']:
                if fixed_ip['ip_address'] == amphora['lb_network_ip']:
                    return port

        # This should never happen
        raise octavia.AmphoraMgmtPortNotFound(
            reason='Could not find the Network Managment Port of'
                   ' amphora {amphora}'.
            format(amphora=amphora['id']))


class HttpRoundRobinAmphoraIpv6Listener(HttpRoundRobinAmphoraIpv4Listener):
    ip_version = 6


class HttpLeastConnectionAmphoraIpv4Listener(
      HttpRoundRobinAmphoraIpv4Listener):
    lb_algorithm = 'LEAST_CONNECTIONS'


class HttpLeastConnectionAmphoraIpv6Listener(
      HttpLeastConnectionAmphoraIpv4Listener):
    ip_version = 6


class HttpSourceIpAmphoraIpv4Listener(HttpRoundRobinAmphoraIpv4Listener):
    lb_algorithm = 'SOURCE_IP'


class HttpSourceIpAmphoraIpv6Listener(HttpSourceIpAmphoraIpv4Listener):
    ip_version = 6
    lb_protocol = 'TCP'


# OVN provider stack fixtures
class OVNIPv4LoadBalancerStack(AmphoraIPv4LoadBalancerStack):
    provider = 'ovn'


class OVNIPv6LoadBalancerStack(OVNIPv4LoadBalancerStack):
    ip_version = 6


class TcpSourceIpPortOvnIpv4Listener(HttpRoundRobinAmphoraIpv4Listener):
    loadbalancer = tobiko.required_fixture(OVNIPv4LoadBalancerStack)
    lb_protocol = 'TCP'
    lb_port = 22
    has_monitor = False
    lb_algorithm = 'SOURCE_IP_PORT'
    pool_protocol = 'TCP'
    application_port = 22


class TcpSourceIpPortOvnIpv6Listener(TcpSourceIpPortOvnIpv4Listener):
    ip_version = 6



# Copyright 2020 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing  # noqa

import tobiko
from tobiko.tests import unit


NOT_CALLED = object()


class TestCached(unit.TobikoUnitTest):

    my_property = tobiko.cached(doc='some doc')

    def test_init(self):
        prop = type(self).my_property
        self.assertIsNone(prop.fget)
        self.assertIsNone(prop.fset)
        self.assertIsNone(prop.fdel)
        self.assertEqual('some doc', prop.__doc__)
        self.assertTrue(prop.cached_id)

    def test_get(self):
        ex = self.assertRaises(AttributeError, lambda: self.my_property)
        self.assertEqual("Cached property has no getter method", str(ex))

    def test_set(self):
        self.my_property = value = object()
        self.assertIs(value, self.my_property)

    def test_delete(self):
        self.my_property = object()
        del self.my_property
        ex = self.assertRaises(AttributeError, lambda: self.my_property)
        self.assertEqual("Cached property has no getter method", str(ex))


class TestCachedWithGetter(unit.TobikoUnitTest):

    @tobiko.cached
    def my_property(self):
        return object()

    def test_init(self):
        # pylint: disable=no-member
        prop = type(self).my_property
        assert isinstance(prop, tobiko.CachedProperty)
        self.assertTrue(callable(prop.fget))
        self.assertIsNone(prop.fset)
        self.assertIsNone(prop.fdel)
        self.assertIs(prop.fget.__doc__, prop.__doc__)
        self.assertTrue(prop.cached_id)

    def test_get(self):
        value = self.my_property
        self.assertIs(value, self.my_property)

    def test_set(self):
        self.my_property = value = object()
        self.assertIs(value, self.my_property)

    def test_delete(self):
        self.my_property = value = object()
        del self.my_property
        self.assertIsNot(value, self.my_property)


class TestCachedWithSetter(unit.TobikoUnitTest):

    _set_value = None

    def set_my_property(self, value):
        self._set_value = value

    my_property = tobiko.cached(fset=set_my_property)

    def test_init(self):
        prop = type(self).my_property
        self.assertIsNone(prop.fget)
        self.assertIs(type(self).set_my_property, prop.fset)
        self.assertIsNone(prop.fdel)
        self.assertIsNone(prop.__doc__)
        self.assertTrue(prop.cached_id)

    def test_get(self):
        self.assertFalse(hasattr(self, 'my_property'))
        self.assertIsNone(self._set_value)

    def test_set(self):
        self.my_property = value = object()
        self.assertIs(value, self.my_property)
        self.assertIs(value, self._set_value)

    def test_delete(self):
        self.my_property = object()
        del self.my_property
        self.assertFalse(hasattr(self, 'my_property'))


class TestCachedWithDeleter(unit.TobikoUnitTest):

    _delete_value = None

    def delete_my_property(self):
        self._delete_value = object()

    my_property = tobiko.cached(fdel=delete_my_property)

    def test_init(self):
        prop = type(self).my_property
        self.assertIsNone(prop.fget)
        self.assertIsNone(prop.fset)
        self.assertIs(type(self).delete_my_property, prop.fdel)
        self.assertIsNone(prop.__doc__)
        self.assertTrue(prop.cached_id)

    def test_get(self):
        self.assertFalse(hasattr(self, 'my_property'))
        self.assertIsNone(self._delete_value)

    def test_set(self):
        self.my_property = value = object()
        self.assertIs(value, self.my_property)
        self.assertIsNone(self._delete_value)

    def test_delete(self):
        self.my_property = object()
        del self.my_property
        self.assertFalse(hasattr(self, 'my_property'))
        self.assertIsNotNone(self._delete_value)


class TestCachedWithDoc(unit.TobikoUnitTest):

    my_doc = 'some doc'

    my_property = tobiko.cached(doc=my_doc)

    @my_property.getter
    def get_my_property(self):
        return object()

    def test_init(self):
        prop = type(self).my_property
        self.assertTrue(callable(prop.fget))
        self.assertIsNone(prop.fset)
        self.assertIsNone(prop.fdel)
        self.assertIs(self.my_doc, prop.__doc__)
        self.assertTrue(prop.cached_id)


class TestCachedWithCachedId(unit.TobikoUnitTest):

    cached_id = 'my_cached_id'

    my_property = tobiko.cached(cached_id=cached_id)

    @my_property.getter
    def get_my_property(self):
        return object()

    def test_init(self):
        prop = type(self).my_property
        self.assertTrue(callable(prop.fget))
        self.assertIsNone(prop.fset)
        self.assertIsNone(prop.fdel)
        self.assertIsNone(prop.__doc__)
        self.assertEqual(self.cached_id, prop.cached_id)
        self.assertFalse(hasattr(self, 'my_cached_id'))

    def test_get(self):
        # pylint: disable=no-member
        value = self.my_property
        self.assertIs(value, self.my_property)
        self.assertIs(value, self.my_cached_id)

    def test_set(self):
        # pylint: disable=no-member
        self.my_property = value = object()
        self.assertIs(value, self.my_property)
        self.assertIs(value, self.my_cached_id)

    def test_delete(self):
        # pylint: disable=no-member
        self.my_property = value = object()
        del self.my_property
        self.assertIsNot(value, self.my_property)
        self.assertIsNot(value, self.my_cached_id)


class TestCachedWithDecorators(unit.TobikoUnitTest):

    my_property = tobiko.cached()

    getter_called = 0

    @my_property.getter
    def getter(self):
        """Getter documentation"""
        self.getter_called += 1
        return object()

    setter_called = tuple()  # type: typing.Tuple

    @my_property.setter
    def setter(self, value):
        self.setter_called += (value,)

    deleter_called = 0

    @my_property.deleter
    def deleter(self):
        self.deleter_called += 1

    def test_init(self):
        # self.my_property method not called yet
        prop = type(self).my_property
        self.assertIs(type(self).getter, prop.fget)
        self.assertIs(type(self).setter, prop.fset)
        self.assertIs(type(self).deleter, prop.fdel)
        self.assertIs(self.getter.__doc__, prop.__doc__)
        self.assertTrue(prop.cached_id)

    def test_get(self):
        # my_property method is called
        prop = type(self).my_property
        value = self.my_property
        self.assertEqual(1, self.getter_called)
        self.assertIs(value, getattr(self, prop.cached_id))
        self.assertEqual((value,), self.setter_called)
        return value

    def test_get_twice(self):
        # my_property method is not called again
        value1 = self.test_get()
        value2 = self.test_get()
        self.assertIs(value1, value2)

    def test_set(self):
        # property value is stored into the object
        prop = type(self).my_property
        self.my_property = value = object()
        self.assertIs(value, self.my_property)
        self.assertEqual(0, self.getter_called)
        self.assertEqual(value, self.setter_called[-1])
        self.assertIs(value, getattr(self, prop.cached_id))
        return value

    def test_set_twice(self):
        # stored value is got after setting
        value1 = self.test_set()
        value2 = self.test_set()
        self.assertIsNot(value1, value2)

    def test_delete(self):
        # deleting remove stored value
        prop = type(self).my_property
        value = self.test_set()
        del self.my_property
        self.assertFalse(hasattr(self, prop.cached_id))
        self.assertIsNot(value, self.my_property)
        self.assertEqual(1, self.deleter_called)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko
from tobiko.shell import sh
from tobiko.openstack import keystone
from tobiko.openstack import nova
from tobiko.openstack import stacks
from tobiko.tests.functional.openstack.stacks import test_cirros


@keystone.skip_unless_has_keystone_credentials()
class CentosServerStackTest(test_cirros.CirrosServerStackTest):
    """Test CentOS server instance"""

    #: Stack of resources with a server attached to a floating IP
    stack = tobiko.required_fixture(stacks.CentosServerStackFixture)

    def test_user_data(self):
        user_data = self.stack.user_data
        self.assertEqual('', user_data)

    def test_python(self):
        python_version = sh.execute(['/usr/libexec/platform-python',
                                     '--version'],
                                    ssh_client=self.stack.ssh_client).stdout
        self.assertTrue(python_version.startswith('Python 3.'),
                        python_version)

    def test_cloud_init_done(self):
        nova.wait_for_cloud_init_done(ssh_client=self.stack.ssh_client)


@tobiko.skip("Can't SSH to CentOS 7 server")
@keystone.skip_unless_has_keystone_credentials()
class Centos7ServerStackTest(CentosServerStackTest):

    #: Stack of resources with a server attached to a floating IP
    stack = tobiko.required_fixture(stacks.Centos7ServerStackFixture)

    def test_python(self):
        python_version = sh.execute(['python', '--version'],
                                    ssh_client=self.stack.ssh_client).stderr
        self.assertTrue(python_version.startswith('Python 2.'),
                        python_version)



# Copyright (c) 2020 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import mock

from tobiko.openstack import tests
from tobiko.tests.unit import openstack


class NeutronAgentTest(openstack.OpenstackTest):

    def setUp(self):
        super(NeutronAgentTest, self).setUp()
        get_neutron_client = self.patch_get_neutron_client()
        self.neutron_client = get_neutron_client.return_value
        self.patch_time()

    def patch_list_agents(self, *args, **kwargs):
        self.neutron_client.list_agents = mock.MagicMock(*args, **kwargs)

    def test_neutron_agents_are_alive_when_healthy(self):
        self.patch_list_agents(return_value=[{'alive': True}])
        agents = tests.test_neutron_agents_are_alive()
        self.assertEqual([{'alive': True}], agents)

    def test_neutron_agents_are_alive_when_no_agents(self):
        self.patch_list_agents(return_value=[])
        agents = tests.test_neutron_agents_are_alive()
        self.assertEqual([], agents)

    def test_neutron_agents_are_alive_when_unhealthy(self):
        self.patch_list_agents(return_value=[{'alive': False}])
        ex = self.assertRaises(self.failureException,
                               tests.test_neutron_agents_are_alive)
        self.assertEqual("Unhealthy agent(s) found:\n"
                         "[\n"
                         "    {\n"
                         '        "alive": false\n'
                         "    }\n"
                         "]\n", str(ex))



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko
from tobiko.shell import sh
from tobiko.tests.functional.openstack import test_topology
from tobiko import tripleo


@tripleo.skip_if_missing_undercloud
class TripleoTopologyTest(test_topology.OpenStackTopologyTest):

    @property
    def topology(self) -> tripleo.TripleoTopology:
        return tobiko.setup_fixture(tripleo.TripleoTopology)

    def test_undercloud_group(self):
        ssh_client = tripleo.undercloud_ssh_client()
        name = sh.get_hostname(ssh_client=ssh_client).split('.')[0]
        node = self.topology.get_node(name)
        self.assertIs(node.ssh_client, ssh_client)
        self.assertEqual(name, node.name)
        nodes = self.topology.get_group('undercloud')
        self.assertEqual([node], nodes)

    @tripleo.skip_if_missing_overcloud
    def test_overcloud_group(self):
        for instance in tripleo.list_overcloud_nodes():
            ssh_client = tripleo.overcloud_ssh_client(instance=instance)
            name = sh.get_hostname(ssh_client=ssh_client).split('.')[0]
            node = self.topology.get_node(name)
            self.assertIs(node.ssh_client, ssh_client)
            self.assertEqual(name, node.name)
            groups = ['overcloud']
            group = name.split('-', 1)[0]
            if group != name:
                groups.append(group)
            for group in groups:
                nodes = self.topology.get_group(group)
                self.assertIn(node, nodes)
                self.assertIn(group, node.groups)
            host_config = tripleo.overcloud_host_config(instance=instance)
            self.assertEqual(host_config.hostname,
                             str(node.public_ip))



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import unittest

import testtools

import tobiko
from tobiko import config
from tobiko.shell import sh
from tobiko.openstack import keystone
from tobiko.openstack import metalsmith
from tobiko import tripleo


CONF = config.CONF


@tripleo.skip_if_missing_undercloud
class UndercloudSshConnectionTest(testtools.TestCase):

    def setUp(self):
        super(UndercloudSshConnectionTest, self).setUp()
        self.ssh_client = tripleo.undercloud_ssh_client()

    def test_connect_to_undercloud(self):
        self.ssh_client.connect()

    def test_fetch_undercloud_credentials(self):
        env = tripleo.load_undercloud_rcfile()
        if not env.get('OS_CLOUD'):
            self.assertTrue(env['OS_AUTH_URL'])
            self.assertTrue(env.get('OS_USERNAME') or env.get('OS_USER_ID'))
            self.assertTrue(env['OS_PASSWORD'])
            self.assertTrue(env.get('OS_TENANT_NAME') or
                            env.get('OS_PROJECT_NAME') or
                            env.get('OS_TENANT_ID') or
                            env.get('OS_PROJECT_ID'))

    def test_execute_command(self):
        result = sh.execute('hostname', ssh_client=self.ssh_client)
        self.assertTrue(result.stdout.startswith('undercloud-0'))


@tripleo.skip_if_missing_undercloud
class UndercloudKeystoneClientTest(testtools.TestCase):

    def test_undercloud_keystone_credentials(self):
        fixture = tripleo.undercloud_keystone_credentials()
        self.assertIsInstance(fixture,
                              keystone.KeystoneCredentialsFixture)
        credentials = keystone.keystone_credentials(fixture)
        credentials.validate()

    def test_undercloud_keystone_session(self):
        session = tripleo.undercloud_keystone_session()
        client = metalsmith.get_metalsmith_client(session=session)
        overcloud_nodes = metalsmith.list_instances(client=client)
        self.assertTrue(overcloud_nodes)

    def test_undercloud_keystone_client(self):
        client = tripleo.undercloud_keystone_client()
        services = keystone.list_services(client=client)
        self.assertTrue(services)


@tripleo.skip_if_missing_undercloud
class UndercloudVersionTest(unittest.TestCase):
    # TODO(eolivare): move the properties to a common class, since they are
    # duplicate in OvercloudVersionTest and UndercloudVersionTest

    @property
    def lower_version(self) -> str:
        v = tripleo.undercloud_version()
        if v.micro > 0:
            lower_v = f"{v.major}.{v.minor}.{v.micro - 1}"
        elif v.minor > 0:
            lower_v = f"{v.major}.{v.minor -1}.{v.micro}"
        elif v.major > 0:
            lower_v = f"{v.major -1}.{v.minor}.{v.micro}"
        else:
            raise ValueError(f"wrong version: {v}")
        return lower_v

    @property
    def same_version(self) -> str:
        v = tripleo.undercloud_version()
        return f"{v.major}.{v.minor}.{v.micro}"

    @property
    def higher_version(self) -> str:
        v = tripleo.undercloud_version()
        return f"{v.major}.{v.minor}.{v.micro + 1}"

    @tripleo.skip_if_missing_undercloud
    def test_undercloud_version(self):
        version = tripleo.undercloud_version()
        self.assertTrue(tobiko.match_version(version, min_version='13'))

    def test_has_undercloud(self):
        self.assertTrue(tripleo.has_undercloud())

    def test_has_undercloud_with_min_version(self):
        self.assertTrue(
            tripleo.has_undercloud(min_version=self.same_version))

    def test_has_undercloud_with_min_version_lower(self):
        self.assertTrue(
            tripleo.has_undercloud(min_version=self.lower_version))

    def test_has_undercloud_with_min_version_higher(self):
        self.assertFalse(
            tripleo.has_undercloud(min_version=self.higher_version))

    def test_has_undercloud_with_max_version(self):
        self.assertFalse(
            tripleo.has_undercloud(max_version=self.same_version))

    def test_has_undercloud_with_max_version_lower(self):
        self.assertFalse(
            tripleo.has_undercloud(max_version=self.lower_version))

    def test_has_undercloud_with_max_version_higher(self):
        self.assertTrue(
            tripleo.has_undercloud(max_version=self.higher_version))

    def test_skip_unless_has_undercloud(self):
        self._assert_test_skip_unless_has_undercloud_dont_skip()

    def test_skip_unless_has_undercloud_with_min_version(self):
        self._assert_test_skip_unless_has_undercloud_dont_skip(
            min_version=self.same_version)

    def test_skip_unless_has_undercloud_with_min_version_lower(self):
        self._assert_test_skip_unless_has_undercloud_dont_skip(
            min_version=self.lower_version)

    def test_skip_unless_has_undercloud_with_min_version_higher(self):
        self._assert_test_skip_unless_has_undercloud_skip(
            min_version=self.higher_version)

    def test_skip_unless_has_undercloud_with_max_version(self):
        self._assert_test_skip_unless_has_undercloud_skip(
            max_version=self.same_version)

    def test_skip_unless_has_undercloud_with_max_version_lower(self):
        self._assert_test_skip_unless_has_undercloud_skip(
            max_version=self.lower_version)

    def test_skip_unless_has_undercloud_with_max_version_higher(self):
        self._assert_test_skip_unless_has_undercloud_dont_skip(
            max_version=self.higher_version)

    def _assert_test_skip_unless_has_undercloud_dont_skip(
            self,
            min_version: tobiko.VersionType = None,
            max_version: tobiko.VersionType = None):
        executed = []

        @tripleo.skip_unless_has_undercloud(min_version=min_version,
                                            max_version=max_version)
        def decorated_function():
            executed.append(True)

        self.assertFalse(executed)
        decorated_function()
        self.assertTrue(executed)

    def _assert_test_skip_unless_has_undercloud_skip(
            self,
            min_version: tobiko.VersionType = None,
            max_version: tobiko.VersionType = None):
        @tripleo.skip_unless_has_undercloud(min_version=min_version,
                                            max_version=max_version)
        def decorated_function():
            raise self.fail('Not skipped')

        with self.assertRaises(unittest.SkipTest):
            decorated_function()



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko
from tobiko import config
from tobiko.openstack import glance
from tobiko.openstack.stacks import _nova


CONF = config.CONF

CENTOS_IMAGE_URL = (
    'https://cloud.centos.org/centos/8/x86_64/images/'
    'CentOS-8-GenericCloud-8.4.2105-20210603.0.x86_64.qcow2')


class CentosImageFixture(glance.URLGlanceImageFixture):
    image_url = CONF.tobiko.centos.image_url or CENTOS_IMAGE_URL
    image_name = CONF.tobiko.centos.image_name
    image_file = CONF.tobiko.centos.image_file
    disk_format = CONF.tobiko.centos.disk_format or "qcow2"
    container_format = CONF.tobiko.centos.container_format or "bare"
    username = CONF.tobiko.centos.username or 'centos'
    password = CONF.tobiko.centos.password
    connection_timeout = CONF.tobiko.centos.connection_timeout or 800.
    disabled_algorithms = CONF.tobiko.centos.disabled_algorithms


CENTOS7_IMAGE_URL = (
    'https://cloud.centos.org/centos/7/images/'
    'CentOS-7-x86_64-GenericCloud.qcow2.xz')


class Centos7ImageFixture(glance.URLGlanceImageFixture):
    image_url = CONF.tobiko.centos7.image_url or CENTOS7_IMAGE_URL
    image_name = CONF.tobiko.centos7.image_name
    image_file = CONF.tobiko.centos7.image_file
    disk_format = CONF.tobiko.centos7.disk_format or "qcow2"
    container_format = CONF.tobiko.centos7.container_format or "bare"
    username = CONF.tobiko.centos7.username or 'centos'
    password = CONF.tobiko.centos7.password
    connection_timeout = CONF.tobiko.centos7.connection_timeout or 800.


class CentosFlavorStackFixture(_nova.FlavorStackFixture):
    ram = 512
    swap = 1024


class CentosServerStackFixture(_nova.CloudInitServerStackFixture):

    #: Glance image used to create a Nova server instance
    image_fixture = tobiko.required_fixture(CentosImageFixture)

    #: Flavor used to create a Nova server instance
    flavor_stack = tobiko.required_fixture(CentosFlavorStackFixture)

    # I expect CentOS based servers to be very slow to boot
    is_reachable_timeout = 900.


class Centos7ServerStackFixture(CentosServerStackFixture):

    #: Glance image used to create a Nova server instance
    image_fixture = tobiko.required_fixture(Centos7ImageFixture)



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import functools

import tobiko
from tobiko.shell import sh


@functools.lru_cache()
def get_rhosp_release(connection: sh.ShellConnectionType = None) \
        -> str:
    connection = sh.shell_connection(connection)
    with connection.open_file('/etc/rhosp-release', 'r') as fd:
        rhosp_release = fd.read().strip()
    if isinstance(rhosp_release, bytes):
        rhosp_release = rhosp_release.decode('UTF-8', 'ignore')
    return rhosp_release


def get_rhosp_version(connection: sh.ShellConnectionType = None) \
        -> tobiko.Version:
    return tobiko.parse_version(get_rhosp_release(connection))



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import shlex

from oslo_log import log
import paramiko
from paramiko import channel

import tobiko
from tobiko.shell.sh import _exception
from tobiko.shell.sh import _execute
from tobiko.shell.sh import _io
from tobiko.shell.sh import _local
from tobiko.shell.sh import _process
from tobiko.shell import ssh
import typing  # noqa

LOG = log.getLogger(__name__)


def ssh_execute(ssh_client, command, environment=None,
                timeout: tobiko.Seconds = None, stdin=None, stdout=None,
                stderr=None, shell=None, expect_exit_status=0, **kwargs):
    """Execute command on remote host using SSH client"""
    process = ssh_process(command=command,
                          environment=environment,
                          timeout=timeout,
                          shell=shell,
                          stdin=stdin,
                          stdout=stdout,
                          stderr=stderr,
                          ssh_client=ssh_client,
                          **kwargs)
    return _execute.execute_process(process=process,
                                    stdin=stdin,
                                    expect_exit_status=expect_exit_status)


def ssh_process(command, environment=None, current_dir=None,
                timeout: tobiko.Seconds = None, shell=None, stdin=None,
                stdout=None, stderr=None, ssh_client=None, sudo=None,
                network_namespace=None):
    if ssh_client is None:
        ssh_client = ssh.ssh_proxy_client()
    if ssh_client:
        return SSHShellProcessFixture(
            command=command, environment=environment, current_dir=current_dir,
            timeout=timeout, shell=shell, stdin=stdin, stdout=stdout,
            stderr=stderr, ssh_client=ssh_client, sudo=sudo,
            network_namespace=network_namespace)
    else:
        return _local.local_process(
            command=command, environment=environment, current_dir=current_dir,
            timeout=timeout, shell=shell, stdin=stdin, stdout=stdout,
            stderr=stderr, sudo=sudo, network_namespace=network_namespace)


class SSHShellProcessParameters(_process.ShellProcessParameters):

    ssh_client = None
    open_session_timeout = 30.0


class SSHShellProcessFixture(_process.ShellProcessFixture):

    def init_parameters(self, **kwargs) -> SSHShellProcessParameters:
        return SSHShellProcessParameters(**kwargs)

    def create_process(self):
        """Execute command on a remote host using SSH client"""
        command = str(self.command)
        ssh_client = self.ssh_client
        parameters = self.parameters

        tobiko.check_valid_type(ssh_client, ssh.SSHClientFixture)
        tobiko.check_valid_type(parameters, SSHShellProcessParameters)
        environment = parameters.environment
        current_dir = parameters.current_dir

        if (hasattr(ssh_client, 'connect_parameters') and
                ssh_client.connect_parameters is not None):
            ssh_client_timeout = ssh_client.connect_parameters.get(
                'connection_timeout')
            ssh_client_attempts = ssh_client.connect_parameters.get(
                'connection_attempts')
        else:
            ssh_client_timeout = None
            ssh_client_attempts = None

        # use ssh_client timeout and attempts values if they are higher than
        # self (SSHShellProcessFixture) values because the values from the
        # outer retry loop should be equal or greater than those from the inner
        # retry loop
        process_retry_timeout = (
            ssh_client_timeout if (
                ssh_client_timeout and (
                    self.parameters.timeout is None or
                    ssh_client_timeout > self.parameters.timeout))
            else self.parameters.timeout)
        process_retry_attempts = (
            ssh_client_attempts if (
                ssh_client_attempts and (
                    self.parameters.retry_count is None or
                    ssh_client_attempts > self.parameters.retry_count))
            else self.parameters.retry_count)

        for attempt in tobiko.retry(
                timeout=process_retry_timeout,
                default_count=process_retry_attempts,
                default_interval=self.parameters.retry_interval,
                default_timeout=self.parameters.retry_timeout):

            timeout = attempt.time_left
            details = (f"command='{command}', "
                       f"current_dir='{current_dir}', "
                       f"login={ssh_client.login}, "
                       f"timeout={timeout}, "
                       f"attempt={attempt}, "
                       f"environment={environment}")
            LOG.debug(f"Create remote process... ({details})")
            try:
                client = ssh_client.connect()
                process = client.get_transport().open_session(
                    timeout=self.open_session_timeout)
                if environment:
                    variables = " ".join(
                        f"{name}={shlex.quote(value)}"
                        for name, value in self.environment.items())
                    command = variables + " " + command
                if current_dir is not None:
                    command = f"cd {current_dir} && {command}"
                process.exec_command(command)
                LOG.debug(f"Remote process created. ({details})")
                return process
            except Exception:
                # Before doing anything else cleanup SSH connection
                ssh_client.close()
                LOG.debug(f"Error creating remote process. ({details})",
                          exc_info=1)
            try:
                attempt.check_limits()
            except tobiko.RetryTimeLimitError as ex:
                LOG.debug(f"Timed out creating remote process. ({details})")
                raise _exception.ShellTimeoutExpired(command=command,
                                                     stdin=None,
                                                     stdout=None,
                                                     stderr=None,
                                                     timeout=timeout) from ex

    def setup_stdin(self):
        self.stdin = _io.ShellStdin(
            delegate=StdinSSHChannelFile(self.process, 'wb'),
            buffer_size=self.parameters.buffer_size)

    def setup_stdout(self):
        self.stdout = _io.ShellStdout(
            delegate=StdoutSSHChannelFile(self.process, 'rb'),
            buffer_size=self.parameters.buffer_size)

    def setup_stderr(self):
        self.stderr = _io.ShellStderr(
            delegate=StderrSSHChannelFile(self.process, 'rb'),
            buffer_size=self.parameters.buffer_size)

    def poll_exit_status(self):
        exit_status = getattr(self.process, 'exit_status', None)
        if exit_status and exit_status < 0:
            exit_status = None
        return exit_status

    def _get_exit_status(self, timeout: tobiko.Seconds = None):
        process = self.process
        if not process.exit_status_ready():
            # Workaround for Paramiko timeout problem
            # CirrOS instances could close SSH channel without sending process
            # exit status
            if timeout is None:
                timeout = 120.
            else:
                timeout = min(timeout, 120.0)
            LOG.debug(f"Waiting for command '{self.command}' exit status "
                      f"(timeout={timeout})")
            # recv_exit_status method doesn't accept timeout parameter
            # therefore here we wait for next channel event expecting it is
            # actually the exit status
            # TODO (fressi): we could use an itimer to set a timeout for
            # recv_exit_status
            if not process.status_event.wait(timeout=timeout):
                LOG.error("Timed out before status event being set "
                          f"(timeout={timeout})")
        if process.exit_status >= 0:
            return process.exit_status
        else:
            return None

    def kill(self, sudo=False):
        process = self.process
        LOG.debug('Killing remote process: %r', self.command)
        try:
            process.close()
        except Exception:
            LOG.exception("Failed killing remote process: %r",
                          self.command)


class SSHChannelFile(channel.ChannelFile):

    def fileno(self):
        return self.channel.fileno()


class StdinSSHChannelFile(SSHChannelFile):

    def close(self):
        super(StdinSSHChannelFile, self).close()
        self.channel.shutdown_write()

    @property
    def write_ready(self):
        return self.channel.send_ready()


class StdoutSSHChannelFile(SSHChannelFile):

    def fileno(self):
        return self.channel.fileno()

    def close(self):
        super(StdoutSSHChannelFile, self).close()
        self.channel.shutdown_read()

    @property
    def read_ready(self):
        return self.channel.recv_ready()


class StderrSSHChannelFile(SSHChannelFile, paramiko.channel.ChannelStderrFile):

    def fileno(self):
        return self.channel.fileno()

    @property
    def read_ready(self):
        return self.channel.recv_stderr_ready()



"""Cache for SSH tunnels."""

from __future__ import absolute_import

import collections
from collections import abc
import getpass
import logging
import os
import subprocess
import threading
import time
import weakref
from contextlib import suppress

import psutil

Context = collections.namedtuple('Context', (
    'uri',
    'interface',
    'local_socket',
    'remote_socket',
    'username',
    'hostname',
    'port',
    'identity_file',
    'ignore_hosts',
    'known_hosts',
))
Context.__new__.__defaults__ = (None, ) * len(Context._fields)  # type: ignore


class Portal(abc.MutableMapping):
    """Expiring container for tunnels."""

    def __init__(self, sweap=25):
        """Construct portal, reap tunnels every sweap seconds."""
        self.data = collections.OrderedDict()
        self.sweap = sweap
        self.ttl = sweap * 2
        self.lock = threading.RLock()
        self._schedule_reaper()

    def __getitem__(self, key):
        """Given uri return tunnel and update TTL."""
        with self.lock:
            value, _ = self.data[key]
            self.data[key] = (value, time.time() + self.ttl)
            self.data.move_to_end(key)
            return value

    def __setitem__(self, key, value):
        """Store given tunnel keyed with uri."""
        if not isinstance(value, Tunnel):
            raise ValueError('Portals only support Tunnels.')

        with self.lock:
            self.data[key] = (value, time.time() + self.ttl)
            self.data.move_to_end(key)

    def __delitem__(self, key):
        """Remove and close tunnel from portal."""
        with self.lock:
            value, _ = self.data[key]
            del self.data[key]
            value.close()
            del value

    def __iter__(self):
        """Iterate tunnels."""
        with self.lock:
            values = list(self.data.values())

        for tunnel, _ in values:
            yield tunnel

    def __len__(self):
        """Return number of tunnels in portal."""
        with self.lock:
            return len(self.data)

    def _schedule_reaper(self):
        timer = threading.Timer(interval=self.sweap, function=self.reap)
        timer.name = 'PortalReaper'
        timer.daemon = True
        timer.start()

    def reap(self):
        """Remove tunnels who's TTL has expired."""
        now = time.time()
        with self.lock:
            reaped_data = self.data.copy()
            for entry in reaped_data.items():
                if entry[1][1] < now:
                    del self.data[entry[0]]
                else:
                    # StopIteration as soon as possible
                    break
            self._schedule_reaper()


class Tunnel:
    """SSH tunnel."""

    def __init__(self, context):
        """Construct Tunnel."""
        self.context = context
        self._closed = True

    @property
    def closed(self):
        """Is tunnel closed."""
        return self._closed

    def bore(self):
        """Create SSH tunnel from given context."""
        cmd = ['ssh', '-fNT']

        if logging.getLogger().getEffectiveLevel() == logging.DEBUG:
            cmd.append('-v')
        else:
            cmd.append('-q')

        if self.context.port:
            cmd.extend(('-p', str(self.context.port)))

        cmd.extend(('-L', '{}:{}'.format(self.context.local_socket,
                                         self.context.remote_socket)))

        if self.context.ignore_hosts:
            cmd.extend(('-o', 'StrictHostKeyChecking=no',
                        '-o', 'UserKnownHostsFile=/dev/null'))
        elif self.context.known_hosts:
            cmd.extend(('-o', 'UserKnownHostsFile={known_hosts}'.format(
                known_hosts=self.context.known_hosts)))

        if self.context.identity_file:
            cmd.extend(('-i', self.context.identity_file))

        cmd.append('{}@{}'.format(self.context.username,
                                  self.context.hostname))

        logging.debug('Opening tunnel "%s", cmd "%s"', self.context.uri,
                      ' '.join(cmd))

        tunnel = subprocess.Popen(cmd, close_fds=True)
        # The return value of Popen() has no long term value as that process
        # has already exited by the time control is returned here. This is a
        # side effect of the -f option. wait() will be called to clean up
        # resources.
        for _ in range(300):
            # TODO: Make timeout configurable
            if os.path.exists(self.context.local_socket) \
                    or tunnel.returncode is not None:
                break
            with suppress(subprocess.TimeoutExpired):
                # waiting for either socket to be created
                # or first child to exit
                tunnel.wait(0.5)
        else:
            raise TimeoutError(
                'Failed to create tunnel "{}", using: "{}"'.format(
                    self.context.uri, ' '.join(cmd)))
        if tunnel.returncode is not None and tunnel.returncode != 0:
            raise subprocess.CalledProcessError(tunnel.returncode,
                                                ' '.join(cmd))
        tunnel.wait()

        self._closed = False
        weakref.finalize(self, self.close)
        return self

    def close(self):
        """Close SSH tunnel."""
        logging.debug('Closing tunnel "%s"', self.context.uri)

        if self._closed:
            return

        # Find all ssh instances for user with uri tunnel the hard way...
        targets = [
            p
            for p in psutil.process_iter(attrs=['name', 'username', 'cmdline'])
            if p.info['username'] == getpass.getuser()
            and p.info['name'] == 'ssh'
            and self.context.local_socket in ' '.join(p.info['cmdline'])
        ]  # yapf: disable

        # ask nicely for ssh to quit, reap results
        for proc in targets:
            proc.terminate()
        _, alive = psutil.wait_procs(targets, timeout=300)

        # kill off the uncooperative, then report any stragglers
        for proc in alive:
            proc.kill()
        _, alive = psutil.wait_procs(targets, timeout=300)

        for proc in alive:
            logging.info('process %d survived SIGKILL, giving up.', proc.pid)

        with suppress(OSError):
            os.remove(self.context.local_socket)
        self._closed = True



# Copyright 2021 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import deprecation

from tobiko import version


def deprecated(deprecated_in: str = None,
               removed_in: str = None,
               details=""):
    return deprecation.deprecated(deprecated_in=deprecated_in,
                                  removed_in=removed_in,
                                  current_version=version.version,
                                  details=details)



# Copyright (c) 2020 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import json

from oslo_log import log

from tobiko.openstack import keystone
from tobiko.openstack.openstackclient import _exception
from tobiko.shell import sh
import tobiko.tripleo


LOG = log.getLogger(__name__)


def execute(cmd, *args, **kwargs):
    arg_list = _param_list(*args, **kwargs)
    cmd_to_exec = cmd.format(params=' '.join(arg_list))
    if tobiko.tripleo.has_undercloud():
        ssh_client = tobiko.tripleo.undercloud_ssh_client()
    else:
        ssh_client = None
    try:
        LOG.debug(f'Command to be executed:\n{cmd_to_exec}')
        result = sh.execute(cmd_to_exec, ssh_client=ssh_client)
    except sh.ShellCommandFailed as ex:
        if ex.exit_status == 1:
            raise _exception.OSPCliApiError(message=f'{ex.stderr}')
        elif ex.exit_status == 2:
            raise _exception.OSPCliClientError(message=f'{ex.stderr}')
        else:
            raise
    output_format = kwargs.pop('format', '')
    if output_format == 'json':
        return json.loads(result.stdout)
    else:
        return dict()


def _param_list(*args, **kwargs):
    if not any(param in kwargs for param in ['os-token', 'os-username']):
        credentials = keystone.keystone_credentials()
        tmp_auth = {}
        tmp_auth['os-auth-url'] = credentials.auth_url
        tmp_auth['os-password'] = credentials.password
        tmp_auth['os-username'] = credentials.username
        tmp_auth['os-cacert'] = credentials.cacert
        tmp_auth['os-project-name'] = credentials.project_name
        tmp_auth['os-user-domain-name'] = credentials.user_domain_name
        tmp_auth['os-project-domain-name'] = credentials.project_domain_name
        tmp_auth['os-project-domain-id'] = credentials.project_domain_id
        if credentials.api_version == 3:
            tmp_auth['os-identity-api-version'] = credentials.api_version
        for key, val in tmp_auth.items():
            if val:
                kwargs[key] = val
    arg_list = []
    for arg in args:
        if len(arg) == 1:
            arg_list.append(f'-{arg}')
        else:
            arg_list.append(f'--{arg}')
    for arg, value in kwargs.items():
        if len(arg) == 1:
            arg_list.append(f'-{arg} {value}')
        else:
            arg_list.append(f'--{arg} {value}')
    return arg_list



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import
from __future__ import division

import typing

import netaddr
from oslo_log import log

import tobiko
from tobiko import config
from tobiko.shell.iperf3 import _execute
from tobiko.shell import ssh


CONF = config.CONF
LOG = log.getLogger(__name__)


def assert_has_bandwith_limits(
        address: typing.Union[str, netaddr.IPAddress],
        min_bandwith: float,
        max_bandwith: float,
        bitrate: int = None,
        download: bool = None,
        port: int = None,
        protocol: str = None,
        ssh_client: ssh.SSHClientType = None,
        timeout: tobiko.Seconds = None) -> None:
    bandwith = _execute.get_bandwidth(address=address,
                                      bitrate=bitrate,
                                      download=download,
                                      port=port,
                                      protocol=protocol,
                                      ssh_client=ssh_client,
                                      timeout=timeout)
    testcase = tobiko.get_test_case()
    LOG.debug(f'measured bandwith: {bandwith}')
    LOG.debug(f'bandwith limits: {min_bandwith} ... {max_bandwith}')
    # an 8% of lower deviation is allowed
    testcase.assertGreater(bandwith, min_bandwith)
    # a 5% of upper deviation is allowed
    testcase.assertLess(bandwith, max_bandwith)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
from collections import abc
import random
import time
import typing

from heatclient.v1 import stacks
from heatclient import exc
from oslo_log import log

import tobiko
from tobiko import config
from tobiko.openstack.heat import _client
from tobiko.openstack.heat import _template
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack import nova


LOG = log.getLogger(__name__)

# Status
INIT_IN_PROGRESS = 'INIT_IN_PROGRESS'
INIT_COMPLETE = 'INIT_COMPLETE'
INIT_FAILED = 'INIT_FAILED'
CREATE_IN_PROGRESS = 'CREATE_IN_PROGRESS'
CREATE_COMPLETE = 'CREATE_COMPLETE'
CREATE_FAILED = 'CREATE_FAILED'
DELETE_IN_PROGRESS = 'DELETE_IN_PROGRESS'
DELETE_COMPLETE = 'DELETE_COMPLETE'
DELETE_FAILED = 'DELETE_FAILED'


TEMPLATE_FILE_SUFFIX = '.yaml'


def heat_stack_parameters(obj,
                          stack: 'HeatStackFixture' = None) \
        -> 'HeatStackParametersFixture':
    if isinstance(obj, HeatStackParametersFixture):
        parameters = obj
    elif obj is None or isinstance(obj, abc.Mapping):
        parameters = HeatStackParametersFixture(stack, obj)
    else:
        parameters = tobiko.get_fixture(obj)
    tobiko.check_valid_type(parameters, HeatStackParametersFixture)
    if stack is not None and parameters.stack is None:
        parameters.stack = stack
    tobiko.check_valid_type(parameters.stack, type(None), HeatStackFixture)
    return parameters


STACK_CLASSES = stacks.Stack,
StackType = typing.Union[stacks.Stack]
StackIdType = typing.Union[str, stacks.Stack, 'HeatStackFixture']


def get_stack_id(stack: StackIdType) -> str:
    if isinstance(stack, str):
        return stack
    elif isinstance(stack, HeatStackFixture):
        return stack.stack_id
    elif isinstance(stack, STACK_CLASSES):
        return stack.id
    else:
        raise TypeError(f'{stack} is not a valid Heat stack ID')


def get_stack(stack: StackIdType,
              resolve_outputs=False,
              client: _client.HeatClientType = None) \
        -> StackType:
    if isinstance(stack, HeatStackFixture):
        return stack.get_stack(resolve_outputs=resolve_outputs)
    else:
        stack_id = get_stack_id(stack)
        return _client.heat_client(client).stacks.get(
            stack_id, resolve_outputs=resolve_outputs)


def list_stacks(client: _client.HeatClientType = None,
                **kwargs) -> tobiko.Selection[StackType]:
    client = _client.heat_client(client)
    return tobiko.select(client.stacks.list(**kwargs))


def find_stack(client: _client.HeatClientType = None,
               unique=False,
               **kwargs) -> StackIdType:
    _stacks = list_stacks(client=client, **kwargs)
    if unique:
        return _stacks.unique
    else:
        return _stacks.first


@keystone.skip_unless_has_keystone_credentials()
class HeatStackFixture(tobiko.SharedFixture):
    """Manages Heat stacks."""

    client: _client.HeatClientType = None
    retry_count: int = 3
    retry_timeout: float = 1200.
    min_retry_interval: float = 0.1
    max_retry_interval: float = 5.
    wait_interval: tobiko.Seconds = 5
    wait_timeout: tobiko.Seconds = 600.
    template: _template.HeatTemplateFixture
    stack: typing.Optional[StackType] = None
    stack_name: typing.Optional[str] = None
    parameters: typing.Optional['HeatStackParametersFixture'] = None
    project: typing.Optional[str] = None
    user: typing.Optional[str] = None

    def __init__(
            self,
            stack_name: str = None,
            template: _template.HeatTemplateFixture = None,
            parameters=None,
            wait_interval: tobiko.Seconds = None,
            client: _client.HeatClientType = None):
        super(HeatStackFixture, self).__init__()
        if stack_name is not None:
            self.stack_name = stack_name
        if template is not None:
            self.template = _template.heat_template(template)
        if parameters is None:
            parameters = self.parameters
        self.parameters = heat_stack_parameters(obj=parameters,
                                                stack=self)
        if client is not None:
            self.client = client
        if wait_interval is not None:
            self.wait_interval = wait_interval

    def setup_fixture(self):
        self.setup_stack_name()
        self.setup_template()
        self.setup_parameters()
        self.setup_client()
        self.setup_project()
        self.setup_user()
        self.setup_stack()

    def setup_template(self):
        tobiko.setup_fixture(self.template)

    def setup_parameters(self):
        self.get_stack_parameters()

    def setup_stack_name(self) -> str:
        stack_name = self.stack_name
        if stack_name is None:
            self.stack_name = stack_name = self.fixture_name
        return stack_name

    def setup_client(self) -> _client.HeatClient:
        client = self.client
        if not isinstance(client, _client.HeatClient):
            self.client = client = _client.heat_client(self.client)
        return client

    @property
    def session(self):
        return self.setup_client().http_client.session

    def setup_project(self):
        if self.project is None:
            self.project = keystone.get_project_id(session=self.session)

    def setup_user(self):
        if self.user is None:
            self.user = keystone.get_user_id(session=self.session)

    def setup_stack(self) -> stacks.Stack:
        stack = self.create_stack()
        tobiko.addme_to_shared_resource(__name__, stack.stack_name)
        return stack

    def get_stack_parameters(self):
        return tobiko.reset_fixture(self.parameters).values

    def create_stack(self, retry: tobiko.Retry = None) -> stacks.Stack:
        if config.get_bool_env('TOBIKO_PREVENT_CREATE'):
            stack = self.validate_created_stack()
        else:
            for attempt in tobiko.retry(retry,
                                        count=self.retry_count,
                                        timeout=self.retry_timeout,
                                        interval=0.):
                try:
                    stack = self.try_create_stack()
                    break
                except InvalidStackError:
                    LOG.exception(f"Error creating stack '{self.stack_name}'",
                                  exc_info=1)
                    if attempt.is_last:
                        raise

                    # the stack shelf counter does not need to be decreased
                    # here, because it was not increased yet
                    self.delete_stack()

                    # It uses a random time sleep to make conflicting
                    # concurrent creations less probable to occur
                    sleep_time = random_sleep_time(
                        min_time=self.min_retry_interval,
                        max_time=self.max_retry_interval)
                    LOG.debug(f"Failed creating stack '{self.stack_name}' "
                              f"(attempt {attempt.number} of "
                              f"{attempt.count}). It will retry after "
                              f"{sleep_time} seconds", exc_info=1)
                    time.sleep(sleep_time)
            else:
                raise RuntimeError('Retry loop broken')
        return stack

    #: valid status expected to be the stack after exiting from create_stack
    # method
    expected_creted_status = {CREATE_IN_PROGRESS, CREATE_COMPLETE}

    def validate_created_stack(self):
        return self.wait_for_stack_status(
            expected_status=self.expected_creted_status,
            check=True)

    def try_create_stack(self) -> stacks.Stack:
        stack = self.wait_for_stack_status(
            expected_status={CREATE_COMPLETE, CREATE_FAILED,
                             CREATE_IN_PROGRESS, DELETE_COMPLETE,
                             DELETE_FAILED})

        if stack is not None:
            stack_status = stack.stack_status
            if stack_status in {CREATE_IN_PROGRESS, CREATE_COMPLETE}:
                LOG.debug(f"Stack already created (name='{self.stack_name}', "
                          f"id='{stack.id}').")
                return self.validate_created_stack()

            if stack_status.endswith('_FAILED'):
                LOG.error(f"Stack '{self.stack_name}' (id='{stack.id}') "
                          f"found in '{stack_status}' status (reason="
                          f"'{stack.stack_status_reason}'). Deleting it...")
                # the stack shelf counter does not need to be decreased here,
                # because it was not increased yet
                self.delete_stack(stack_id=stack.id)

            self.wait_until_stack_deleted()

        # Cleanup cached objects
        assert self.stack is None
        self._outputs = self._resources = None

        # Re-compile template parameters
        parameters = self.get_stack_parameters()

        # Ensure quota limits are OK just in time before start creating
        # a new stack
        self.ensure_quota_limits()

        self.prepare_external_resources()

        LOG.debug('Begin creating stack %r...', self.stack_name)
        try:
            stack_id: str = self.setup_client().stacks.create(
                stack_name=self.stack_name,
                template=self.template.template_yaml,
                parameters=parameters)['stack']['id']
        except exc.HTTPConflict:
            LOG.debug(f"Stack '{self.stack_name}' already created")
            return self.validate_created_stack()

        LOG.debug(f"New stack being created: name='{self.stack_name}', "
                  f"id='{stack_id}'.")
        try:
            stack = self.validate_created_stack()
        except InvalidStackError as ex:
            LOG.debug(f'Deleting invalid stack (name={self.stack_name}, "'
                      f'"id={stack_id}): {ex}')
            # the stack shelf counter does not need to be decreased here,
            # because it was not increased yet
            self.delete_stack(stack_id=stack_id)
            raise

        if stack_id != stack.id:
            LOG.debug(f'Deleting duplicate stack (name={self.stack_name}, "'
                      f'"id={stack_id})')
            # the stack shelf counter does not need to be decreased here,
            # because it was not increased yet
            self.delete_stack(stack_id=stack_id)
            del stack_id

        LOG.debug('Stack successfully created "'
                  f"(name={self.stack_name}, id={stack.id}).")
        return stack

    def prepare_external_resources(self):
        pass

    _resources = None

    @tobiko.fixture_property
    def resources(self):
        resources = self._resources
        if not self._resources:
            self._resources = resources = HeatStackResourceFixture(self)
        return resources

    def cleanup_fixture(self):
        n_tests_using_stack = len(tobiko.removeme_from_shared_resource(
            __name__, self.stack_name))
        if n_tests_using_stack == 0:
            self.setup_client()
            self.cleanup_stack()
        else:
            LOG.info('Stack %r not deleted because %d tests are using it',
                     self.stack_name, n_tests_using_stack)

    def cleanup_stack(self):
        self.delete_stack()
        self.wait_until_stack_deleted()

    def delete_stack(self, stack_id=None):
        """Deletes stack."""
        self.stack = self._outputs = self._resources = None
        if not stack_id:
            stack_id = self.stack_id
        try:
            self.client.stacks.delete(stack_id)
        except exc.NotFound:
            LOG.debug('Stack already deleted: %r (id=%r)', self.stack_name,
                      stack_id)
        else:
            LOG.debug('Deleting stack %r (id=%r)...', self.stack_name,
                      stack_id)

    @property
    def stack_id(self) -> str:
        stack = self.stack
        if stack is None:
            return self.setup_stack_name()
        else:
            return stack.id

    def get_stack(self, resolve_outputs=False) \
            -> typing.Optional[stacks.Stack]:
        """Returns stack ID."""
        client = self.setup_client()
        try:
            self.stack = stack = client.stacks.get(
                self.stack_name, resolve_outputs=resolve_outputs)
        except exc.HTTPNotFound:
            LOG.debug(f"Stack '{self.stack_name}' not found")
            self.stack = stack = None
        finally:
            self._outputs = self._resources = None
        return stack

    def wait_for_create_complete(self,
                                 cached=True,
                                 check=True,
                                 timeout: tobiko.Seconds = None,
                                 interval: tobiko.Seconds = None) \
            -> typing.Optional[stacks.Stack]:
        return self.wait_for_stack_status(expected_status={CREATE_COMPLETE},
                                          cached=cached,
                                          check=check,
                                          timeout=timeout,
                                          interval=interval)

    def wait_for_delete_complete(self,
                                 check=True,
                                 cached=True,
                                 timeout: tobiko.Seconds = None,
                                 interval: tobiko.Seconds = None) \
            -> typing.Optional[stacks.Stack]:
        return self.wait_for_stack_status(expected_status={DELETE_COMPLETE},
                                          cached=cached,
                                          check=check,
                                          timeout=timeout,
                                          interval=interval)

    def wait_until_stack_deleted(self,
                                 check=True,
                                 cached=True,
                                 timeout: tobiko.Seconds = None,
                                 interval: tobiko.Seconds = None):
        # check stack has been completely deleted
        for attempt in tobiko.retry(timeout=timeout,
                                    interval=interval,
                                    default_timeout=self.wait_timeout,
                                    default_interval=self.wait_interval):
            # Ensure to refresh stack status
            stack = self.wait_for_delete_complete(check=check,
                                                  cached=cached,
                                                  timeout=attempt.time_left,
                                                  interval=attempt.interval)
            if stack is None:
                LOG.debug(f"Stack {self.stack_name} disappeared")
                break

            assert stack.stack_status == DELETE_COMPLETE
            if attempt.is_last:
                raise HeatStackDeletionFailed(
                    name=self.stack_name,
                    observed=stack.stack_status,
                    expected={DELETE_COMPLETE},
                    status_reason=stack.stack_status_reason)

            cached = False
            LOG.debug("Waiting for deleted stack to disappear: '%s'",
                      self.stack_name)
        else:
            raise RuntimeError("Retry look broken itself")

    def wait_for_stack_status(
            self,
            expected_status: typing.Container[str],
            check=True,
            cached=True,
            timeout: tobiko.Seconds = None,
            interval: tobiko.Seconds = None) \
            -> typing.Optional[stacks.Stack]:
        """Waits for the stack to reach the given status."""
        for attempt in tobiko.retry(
                timeout=timeout,
                interval=interval,
                default_timeout=self.wait_timeout,
                default_interval=self.wait_interval):
            if cached:
                cached = False
                stack = self.stack or self.get_stack()
            else:
                stack = self.get_stack()
            stack_status = getattr(stack, 'stack_status', DELETE_COMPLETE)
            if stack_status in expected_status:
                LOG.debug(f"Stack '{self.stack_name}' reached expected "
                          f"status: '{stack_status}'")
                break

            if not stack_status.endswith('_IN_PROGRESS'):
                LOG.warning(f"Stack '{self.stack_name}' reached unexpected "
                            f"status: '{stack_status}'")
                break

            if attempt.is_last:
                LOG.warning(f"Timed out waiting for stack '{self.stack_name}' "
                            f"status to change from '{stack_status}' to "
                            f"'{expected_status}'.")
                break

            LOG.debug(f"Waiting for stack '{self.stack_name}' status to "
                      f"change from '{stack_status}' to "
                      f"'{expected_status}'...")
        else:
            raise RuntimeError('Retry loop broken')

        if stack is not None:
            self._log_stack_status(stack)

        if check:
            if stack is None:
                if DELETE_COMPLETE not in expected_status:
                    raise HeatStackNotFound(name=self.stack_name)
            else:
                check_stack_status(stack, expected_status)

        return stack

    _outputs = None

    def _log_stack_status(self, stack):
        if stack is None:
            LOG.debug(f"Stack '{self.stack_name}' doesn't exist")
        else:
            stack_status_reason = stack.stack_status_reason
            if stack_status_reason:
                LOG.info(f"Stack '{self.stack_name}' status is "
                         f"'{stack.stack_status}'. Reason:\n"
                         f"\t{stack_status_reason}")
            else:
                LOG.debug(f"Stack '{self.stack_name}' status is "
                          f"'{stack.stack_status}'.")

    def get_stack_outputs(self):
        outputs = self._outputs
        if not outputs:
            self._outputs = outputs = HeatStackOutputsFixture(self)
        return outputs

    outputs = tobiko.fixture_property(get_stack_outputs)

    def __getattr__(self, name):
        try:
            return self.get_stack_outputs().get_value(name)
        except HeatStackOutputKeyError:
            pass
        message = "Object {!r} has no attribute {!r}".format(self, name)
        raise AttributeError(message)

    def ensure_quota_limits(self):
        """Ensures quota limits before creating a new stack
        """
        try:
            self.ensure_neutron_quota_limits()
            self.ensure_nova_quota_limits()
        except (nova.EnsureNovaQuotaLimitsError,
                neutron.EnsureNeutronQuotaLimitsError) as ex:
            raise InvalidStackError(name=self.stack_name) from ex

    def ensure_neutron_quota_limits(self):
        required_quota_set = self.neutron_required_quota_set
        if required_quota_set:
            neutron.ensure_neutron_quota_limits(project=self.project,
                                                **required_quota_set)

    def ensure_nova_quota_limits(self):
        required_quota_set = self.nova_required_quota_set
        if required_quota_set:
            nova.ensure_nova_quota_limits(project=self.project,
                                          user=self.user,
                                          **required_quota_set)

    @property
    def neutron_required_quota_set(self) -> typing.Dict[str, int]:
        return collections.defaultdict(int)

    @property
    def nova_required_quota_set(self) -> typing.Dict[str, int]:
        return collections.defaultdict(int)


class HeatStackKeyError(tobiko.TobikoException):
    message = "key {key!r} not found in stack {name!r}"


class HeatStackResourceKeyError(HeatStackKeyError):
    message = "resource key {key!r} not found in stack {name!r}"


class HeatStackParameterKeyError(HeatStackKeyError):
    message = "parameter key {key!r} not found in stack {name!r}"


class HeatStackOutputKeyError(HeatStackKeyError):
    message = "output key {key!r} not found in stack {name!r}"


class HeatStackNamespaceFixture(tobiko.SharedFixture):

    key_error = HeatStackKeyError
    _keys = None
    _values = None

    def __init__(self, stack):
        super(HeatStackNamespaceFixture, self).__init__()
        if stack and not isinstance(stack, HeatStackFixture):
            message = "Object {!r} is not an HeatStackFixture".format(stack)
            raise TypeError(message)
        self.stack = stack

    def setup_fixture(self):
        self.setup_keys()
        self.setup_values()

    def setup_keys(self):
        keys = self._keys
        if keys is None:
            self._keys = keys = self.get_keys()
            self.addCleanup(self.cleanup_keys)
        return keys

    keys = tobiko.fixture_property(setup_keys)

    def get_keys(self):
        raise NotImplementedError

    def cleanup_keys(self):
        del self._keys

    def setup_values(self):
        values = self._values
        if values is None:
            self._values = values = self.get_values()
            self.addCleanup(self.cleanup_values)
        return values

    values = tobiko.fixture_property(setup_values)

    def get_values(self):
        raise NotImplementedError

    def cleanup_values(self):
        del self._values

    def get_value(self, key):
        # Match template outputs definition before getting value
        if key in self.keys:
            try:
                return self.values[key]
            except KeyError:
                LOG.error('Key %r not found in stack %r', key,
                          self.stack.stack_name)
        else:
            LOG.error('Key %r not found in template for stack %r', key,
                      self.stack.stack_name)
        raise self.key_error(name=self.stack.stack_name, key=key)

    def set_value(self, key, value):
        # Match template outputs definition before setting value
        if key in self.keys:
            self.values[key] = value
        else:
            LOG.error('Key %r not found in template for stack %r', key,
                      self.stack.stack_name)
        raise self.key_error(name=self.stack.stack_name, key=key)

    def __getattr__(self, name):
        try:
            return self.get_value(name)
        except self.key_error:
            pass
        message = "Object {!r} has no attribute {!r}".format(self, name)
        raise AttributeError(message)


class HeatStackParametersFixture(HeatStackNamespaceFixture):

    key_error = HeatStackParameterKeyError

    def __init__(self, stack, parameters=None):
        super(HeatStackParametersFixture, self).__init__(stack)
        self.parameters = parameters and dict(parameters) or {}

    def get_keys(self):
        template = tobiko.setup_fixture(self.stack.template)
        return frozenset(template.parameters or [])

    def get_values(self):
        values = dict(self.parameters)
        missing_keys = sorted(self.keys - set(values))
        for key in missing_keys:
            value = getattr(self.stack, key, None)
            if value is not None:
                values[key] = value
        return values


class HeatStackOutputsFixture(HeatStackNamespaceFixture):

    key_error = HeatStackOutputKeyError

    def get_keys(self):
        template = tobiko.setup_fixture(self.stack.template)
        return frozenset(template.outputs or [])

    def get_values(self):
        # Can't get output values before stack creation is complete
        self.stack.wait_for_create_complete()
        outputs = self.stack.get_stack(resolve_outputs=True).outputs
        return {o['output_key']: o['output_value']
                for o in outputs}


def check_stack_status(stack: stacks.Stack,
                       expected_status: typing.Container[str]):
    stack_status = stack.stack_status
    if stack_status in expected_status:
        return stack_status
    if stack_status == CREATE_FAILED and (
            CREATE_IN_PROGRESS in expected_status or
            CREATE_COMPLETE in expected_status):
        raise HeatStackCreationFailed(
            name=stack.stack_name,
            observed=stack_status,
            expected=expected_status,
            status_reason=stack.stack_status_reason)
    if stack_status == DELETE_FAILED and (
            DELETE_IN_PROGRESS in expected_status or
            DELETE_COMPLETE in expected_status):
        raise HeatStackDeletionFailed(
            name=stack.stack_name,
            observed=stack_status,
            expected=expected_status,
            status_reason=stack.stack_status_reason)
    raise InvalidHeatStackStatus(
        name=stack.stack_name,
        observed=stack_status,
        expected=expected_status,
        status_reason=stack.stack_status_reason)


class HeatStackError(tobiko.TobikoException):
    message = "error creating stack '{name}'"


class HeatStackNotFound(HeatStackError):
    message = "stack {name!r} not found"


class InvalidStackError(HeatStackError):
    message = "invalid stack {name!r}"


class InvalidHeatStackStatus(InvalidStackError):
    message = ("stack {name!r} status {observed!r} not in {expected!r}\n"
               "{status_reason!s}")


class HeatStackCreationFailed(InvalidHeatStackStatus):
    pass


class HeatStackDeletionFailed(InvalidHeatStackStatus):
    pass


class HeatStackResourceFixture(HeatStackNamespaceFixture):

    key_error = HeatStackResourceKeyError

    def get_keys(self):
        template = tobiko.setup_fixture(self.stack.template)
        return frozenset(template.resources or [])

    def get_values(self):
        self.stack.wait_for_create_complete()
        client = self.stack.client
        resources = client.resources.list(self.stack.stack_id)
        return {r.resource_name: r for r in resources}

    @property
    def fixture_name(self):
        return self.stack_name + '.resources'


def random_sleep_time(min_time, max_time):
    assert min_time <= min_time
    return (max_time - min_time) * random.random() + min_time



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os

import testtools

import tobiko
from tobiko import config
from tobiko.openstack import keystone
from tobiko.openstack import stacks
from tobiko.shell import sh
from tobiko.shell import ssh


CONF = config.CONF


class WhichTest(testtools.TestCase):

    @property
    def ssh_client(self) -> ssh.SSHClientType:
        return False

    def test_find_command(self):
        result = sh.find_command(command='which',
                                 ssh_client=self.ssh_client)
        self.assertEqual('which', os.path.basename(result))

    def test_find_command_with_invalid(self):
        ex = self.assertRaises(sh.CommandNotFound,
                               sh.find_command,
                               command='<invalid_command>',
                               ssh_client=self.ssh_client)
        self.assertEqual('<invalid_command>', ex.command)
        self.assertEqual(sh.get_hostname(ssh_client=self.ssh_client),
                         ex.hostname)
        self.assertIsInstance(ex.__cause__, sh.ShellCommandFailed)

    def test_find_command_with_skip(self):
        result = sh.find_command(command='which',
                                 ssh_client=self.ssh_client,
                                 skip=True)
        self.assertEqual('which', os.path.basename(result))

    def test_find_command_with_invalid_and_skip(self):
        ex = self.assertRaises(sh.SkipOnCommandNotFound,
                               sh.find_command,
                               command='<invalid_command>',
                               skip=True,
                               ssh_client=self.ssh_client)
        self.assertIsInstance(ex, tobiko.SkipException)
        self.assertIsInstance(ex, sh.CommandNotFound)
        self.assertEqual('<invalid_command>', ex.command)
        self.assertEqual(sh.get_hostname(ssh_client=self.ssh_client),
                         ex.hostname)
        self.assertIsInstance(ex.__cause__, sh.ShellCommandFailed)


class ProxyJumpWhichTest(WhichTest):

    def setUp(self):
        super().setUp()
        if ssh.ssh_proxy_client() is None:
            self.skipTest('SSH proxy jump not configured')

    @property
    def ssh_client(self) -> ssh.SSHClientType:
        return None


@keystone.skip_unless_has_keystone_credentials()
class SSHWhichTest(WhichTest):

    server_stack = tobiko.required_fixture(
        stacks.UbuntuMinimalServerStackFixture)

    @property
    def ssh_client(self):
        return self.server_stack.ssh_client


@keystone.skip_unless_has_keystone_credentials()
class CirrosExecuteTest(SSHWhichTest):
    server_stack = tobiko.required_fixture(stacks.CirrosServerStackFixture)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import io
import select

from oslo_log import log


LOG = log.getLogger(__name__)


class ShellIOBase(io.IOBase):

    buffer_size = io.DEFAULT_BUFFER_SIZE

    def __init__(self, delegate, fd=None, buffer_size=None):
        super(ShellIOBase, self).__init__()
        self.delegate = delegate
        if buffer_size:
            self.buffer_size = int(buffer_size)
        if fd is None:
            fd = delegate.fileno()
        self.fd = fd
        self._data_chunks = []

    @property
    def data(self):
        chunks = self._data_chunks
        if not chunks:
            return None

        chunks_number = len(chunks)
        if chunks_number == 1:
            return chunks[0]

        # Use a zero-length object of chunk type to join chunks
        data = chunks[0][:0].join(chunks)
        self._data_chunks = chunks = [data]
        return data

    def __str__(self):
        data = self.data
        if not data:
            return ''

        if isinstance(data, str):
            return data

        return data.decode()

    def fileno(self):
        return self.fd

    def readable(self):
        return False

    def writable(self):
        return False

    def close(self):
        self.delegate.close()

    @property
    def closed(self):
        return self.delegate.closed


class ShellReadable(ShellIOBase):

    def readable(self):
        return True

    def read(self, size: int = None) -> bytes:
        size = size or self.buffer_size
        try:
            chunk: bytes = self.delegate.read(size) or b''
        except IOError:
            LOG.exception('Error reading from %r', self)
            try:
                self.close()
            except Exception:
                LOG.exception('Error closing %r', self)
            raise

        if chunk:
            self._data_chunks.append(chunk)
        return chunk

    @property
    def read_ready(self):
        return (not self.closed and
                getattr(self.delegate, 'read_ready', False))

    def readall(self, size=None):
        return join_chunks(self._readall(size))

    def _readall(self, size):
        while self.read_ready:
            chunk = self.read(size=size)
            if chunk:
                yield chunk
            else:
                break


class ShellWritable(ShellIOBase):

    def writable(self):
        return True

    def write(self, data):
        if not isinstance(data, bytes):
            data = data.encode()
        witten_bytes = self.delegate.write(data)
        if witten_bytes is None:
            witten_bytes = len(data)
        self._data_chunks.append(data)
        return witten_bytes

    @property
    def write_ready(self):
        return (not self.closed and
                getattr(self.delegate, 'write_ready', False))


class ShellStdin(ShellWritable):
    pass


class ShellStdout(ShellReadable):
    pass


class ShellStderr(ShellReadable):
    pass


def select_files(files, timeout, mode='rw'):
    # NOTE: in case there is no files that can be selected for given mode,
    # this function is going to behave like time.sleep()
    if timeout is None:
        message = "Invalid value for timeout: {!r}".format(timeout)
        raise ValueError(message)

    timeout = float(timeout)
    opened = select_opened_files(files)
    readable = writable = set()
    if 'r' in mode:
        readable = select_readable_files(opened)
    if 'w' in mode:
        writable = select_writable_files(opened)

    read_ready = select_read_ready_files(readable)
    write_ready = select_write_ready_files(writable)
    if not write_ready and not read_ready:
        rlist, wlist, xlist = select.select(readable, writable, opened,
                                            timeout)
        read_ready = readable & set(rlist + xlist)
        write_ready = writable & set(wlist + xlist)

    return read_ready, write_ready


def select_opened_files(files):
    return {f for f in files if is_opened_file(f)}


def is_opened_file(f):
    return not getattr(f, 'closed', True)


def select_readable_files(files):
    return {f for f in files if is_readable_file(f)}


def is_readable_file(f):
    return f.readable()


def select_read_ready_files(files):
    return {f for f in files if f.read_ready}


def select_writable_files(files):
    return {f for f in files if is_writable_file(f)}


def is_writable_file(f):
    return f.writable()


def select_write_ready_files(files):
    return {f for f in files if f.write_ready}


def join_chunks(chunks):
    chunk_it = iter(chunks)
    data = None
    for chunk in chunk_it:
        if chunk:
            data = chunk
            break
    if data:
        # Use a zero-length chunk to join other chunks
        return data + data[:0].join(chunk for chunk in chunk_it if chunk)
    else:
        return None



from __future__ import absolute_import

import collections
import json
import re
import typing

from keystoneauth1 import exceptions
import netaddr
from oslo_log import log

import tobiko
from tobiko.openstack import neutron
from tobiko.openstack import topology
from tobiko.shell import ip
from tobiko.shell import sh
from tobiko.shell import ss
from tobiko.tripleo import _overcloud
from tobiko.tripleo import pacemaker

LOG = log.getLogger(__name__)


# Supported OVN DB service models
# Supported OVN databases
OVNDBS = ('nb', 'sb')
DBNAMES = {'nb': 'OVN_Northbound', 'sb': 'OVN_Southbound'}


def build_ovn_db_show_dict(ovn_db_show_str):
    # returns a dictionary with OVN NB or OVN SB DB information
    # each dict key is a section from the OVN DB command output
    # each dict value is a list of strings
    ovn_master_db_dict = {}
    for line in ovn_db_show_str.splitlines():
        if not re.match(r'^\s+', line):
            # if line does not start with a space char, a new section will be
            # started, i.e. a new dict entry
            current_ovn_section = line.strip()
            ovn_master_db_dict[current_ovn_section] = []
        else:
            ovn_master_db_dict[current_ovn_section].append(line.strip())

    for section in ovn_master_db_dict:
        ovn_master_db_dict[section].sort()
    return ovn_master_db_dict


def test_neutron_agents_are_alive(timeout=300., interval=5.) \
        -> tobiko.Selection[neutron.NeutronAgentType]:
    for attempt in tobiko.retry(timeout=timeout, interval=interval):
        LOG.debug("Look for unhealthy Neutron agents...")
        try:
            # get Neutron agent list
            agents = neutron.list_agents()
        except (neutron.ServiceUnavailable,
                neutron.NeutronClientException,
                exceptions.connection.ConnectFailure) as ex:
            if attempt.is_last:
                raise
            else:
                # retry because Neutron server could still be unavailable
                # after a disruption
                LOG.debug(f"Waiting for neutron service... ({ex})")
                continue  # Let retry

        dead_agents = agents.with_items(alive=False)
        if dead_agents:
            dead_agents_details = json.dumps(agents, indent=4, sort_keys=True)
            if attempt.is_last:
                tobiko.fail("Unhealthy agent(s) found:\n"
                            f"{dead_agents_details}\n")
            else:
                # retry because some Neutron agent could still be unavailable
                # after a disruption
                LOG.debug("Waiting for Neutron agents to get alive...\n"
                          f"{dead_agents_details}")
                continue

        LOG.debug(f"All {len(agents)} Neutron agents are alive.")
        break
    else:
        raise RuntimeError("Retry loop broken")

    return agents


def ovn_dbs_vip_bindings(test_case):
    ovn_conn_str = get_ovn_db_connections()
    # ovn db sockets might be centrillized or distributed
    # that depends on the openstack version under test
    sockets_centrallized = topology.verify_osp_version('14.0', lower=True)
    for db in OVNDBS:
        found_centralized = False
        addrs, port = parse_ips_from_db_connections(ovn_conn_str[db])
        if _overcloud.is_ovn_using_raft():
            addrs.append(netaddr.IPAddress('0.0.0.0'))
            addrs.append(netaddr.IPAddress('::'))
        for node in topology.list_openstack_nodes(group='controller'):
            socs = ss.tcp_listening(port=port, ssh_client=node.ssh_client)
            if sockets_centrallized and not socs:
                continue
            test_case.assertEqual(1, len(socs))
            test_case.assertIn(socs[0]['local_addr'], addrs)
            test_case.assertEqual(socs[0]['process'][0], 'ovsdb-server')
            if sockets_centrallized:
                test_case.assertFalse(found_centralized)
                found_centralized = True
        if sockets_centrallized:
            test_case.assertTrue(found_centralized)


def get_ovn_db_connections():
    """Fetches OVN DB connection strings from ml2_conf.ini file

    In HA environments there is a single IP address in the connection
    string <PROTO>:<ADDR>:<PORT>. There is always cluster virtual IP
    address.
        tcp:172.17.1.95:6641
        ssl:172.17.1.95:6641
        tcp:[fd00:fd00:fd00:2000::356]:6641

    In RAFT environments there are separate IP address for each cluster
    member divided by comma. Protocols and ports should be similar in such
    a configuration.
        tcp:172.17.1.24:6641,tcp:172.17.1.90:6641,tcp:172.17.1.115:6641
    """
    ml2_conf = topology.get_config_file_path('ml2_conf.ini')
    ctl_ssh = topology.list_openstack_nodes(group='controller')[0].ssh_client
    con_strs = {}
    for db in OVNDBS:
        cmd = 'crudini --get {} ovn ovn_{}_connection'.format(ml2_conf, db)
        output = sh.execute(cmd, ssh_client=ctl_ssh, sudo=True).stdout
        con_strs[db] = output.splitlines()[0]
    LOG.debug('OVN DB connection string fetched from {} file: {}'.format(
        ml2_conf, con_strs))
    return con_strs


class InvalidDBConnString(tobiko.TobikoException):
    pass


def parse_ips_from_db_connections(con_str):
    """Parse OVN DB connection string to get IP addresses

    In HA environments there is a single IP address in the connection
    string <PROTO>:<ADDR>:<PORT>. There is always cluster virtual IP
    address.
        tcp:172.17.1.95:6641
        ssl:172.17.1.95:6641
        tcp:[fd00:fd00:fd00:2000::356]:6641

    In RAFT environments there are separate IP address for each cluster
    member divided by comma. Protocols and ports should be similar in such
    a configuration.
        tcp:172.17.1.24:6641,tcp:172.17.1.90:6641,tcp:172.17.1.115:6641
    """
    addrs = []
    ref_port = ''
    ref_protocol = ''
    for substr in con_str.split(','):
        try:
            protocol, con = substr.split(':', 1)
            tmp_addr, port = con.rsplit(':', 1)
        except (ValueError, AttributeError) as ex:
            msg = 'Fail to parse "{}" substring of "{}" OVN DB connection '\
                    'string'.format(substr, con_str)
            LOG.error(msg)
            raise InvalidDBConnString(message=msg) from ex
        if not ref_port:
            ref_port = port
        if not ref_protocol:
            ref_protocol = protocol
        if protocol != ref_protocol or port != ref_port:
            msg = 'Ports or protocols are not identical for OVN DB'\
                    'connections: {}'.format(con_str)
            LOG.error(msg)
            raise InvalidDBConnString(message=msg)
        try:
            addr = netaddr.IPAddress(tmp_addr.strip(']['))
        except ValueError as ex:
            msg = 'Invalid IP address "{}" in "{}"'.format(addr, con_str)
            LOG.error(msg)
            raise InvalidDBConnString(message=msg) from ex
        addrs.append(addr)
    LOG.debug('Addresses are parsed from OVN DB connection string: {}'.format(
        addrs))
    return addrs, ref_port


def ovn_dbs_are_synchronized(test_case):
    """Check that OVN DBs are syncronized across all controller nodes"""
    db_sync_status = get_ovn_db_sync_status()
    if _overcloud.is_ovn_using_ha():
        # In Active-Backup service model we expect the same controller to be
        # active for both databases. This controller node should be configured
        # for virtual IP in pacemaker. Other controllers should be in backup
        # state
        ovn_master_node_name = pacemaker.get_ovn_db_master_node()
        test_case.assertEqual(1, len(ovn_master_node_name))
        ovn_master_node = topology.get_openstack_node(ovn_master_node_name[0])
        LOG.debug("OVN DB master node hostname is: {}".format(
            ovn_master_node.hostname))
        for db in OVNDBS:
            for controller, state in db_sync_status[db]:
                if controller == ovn_master_node.hostname:
                    test_case.assertEqual('active', state)
                else:
                    test_case.assertEqual('backup', state)
    elif _overcloud.is_ovn_using_raft():
        # In clustered database service model we expect all databases to be
        # active
        for db in OVNDBS:
            for _, state in db_sync_status[db]:
                test_case.assertEqual('active', state)
    dumps = dump_ovn_databases()
    for db in OVNDBS:
        if len(dumps[db]) <= 1:
            # Database from a single node is available
            # so there is nothing to compare it with
            continue
        for i in range(1, len(dumps[db])):
            test_case.assertEqual(dumps[db][0][1], dumps[db][i][1])
            LOG.debug('OVN {} databases are equal on {} and {}'.format(
                db, dumps[db][0][0], dumps[db][i][0]))


def find_ovn_db_sockets():
    """Search for OVN DB sockets

    Unix sockets are useful in case there is a need to check the local
    database.
    """
    node_ssh = topology.list_openstack_nodes(group='controller')[0].ssh_client
    sockets = {}
    for db in OVNDBS:
        db_socket = ss.unix_listening(file_name='*ovn{}_db.sock'.format(db),
                                      ssh_client=node_ssh)
        if len(db_socket) < 1 or not db_socket[0].get('local_addr'):
            msg = 'Failed to find active unix socket for {}'.format(db)
            LOG.error(msg)
            raise ss.SocketLookupError(message=msg)
        sockets[db] = db_socket[0]['local_addr']
    LOG.debug('OVN DB socket files found: {}'.format(sockets))
    return sockets


def dump_ovn_databases():
    """Dump NB and SB on each controller node"""
    from tobiko.tripleo import containers
    runtime_name = containers.get_container_runtime_name()
    sockets = find_ovn_db_sockets()
    # To be able to connect to local database in RAFT environment
    # --no-leader-only parameter should be specified
    no_leader = (' --no-leader-only' if _overcloud.is_ovn_using_raft() else '')
    dumps = {}
    for node in topology.list_openstack_nodes(group='controller'):
        for db in OVNDBS:
            connection = '--db=unix:{}{}'.format(sockets[db], no_leader)
            cmd = '{} exec -uroot ovn_controller ovn-{}ctl {} show'.format(
                    runtime_name, db, connection)
            LOG.debug('Dump {} database on {} with following command: {}'.
                      format(db, node.hostname, cmd))
            output = sh.execute(cmd, ssh_client=node.ssh_client, sudo=True)
            dumps.setdefault(db, [])
            dumps[db].append(
                    [node.hostname, build_ovn_db_show_dict(output.stdout)])
    return dumps


def find_ovn_db_ctl_files():
    """Search for ovnsb_db.ctl and ovnnb_db.ctl files"""
    node = topology.list_openstack_nodes(group='controller')[0]
    ctl_files = {}
    for db in OVNDBS:
        cmd = 'find /var/ -name ovn{}_db.ctl'.format(db)
        found = sh.execute(cmd, ssh_client=node.ssh_client, sudo=True).stdout
        ctl_files[db] = found.strip()
    LOG.debug('OVN DB ctl files found: {}'.format(ctl_files))
    return ctl_files


def get_ovn_db_sync_status():
    """Query sync status for NB and SB for each controller node"""
    db_sync_status = {}
    ctl_files = find_ovn_db_ctl_files()
    for node in topology.list_openstack_nodes(group='controller'):
        for db in OVNDBS:
            ctl_file = ctl_files[db]
            cmd = 'ovs-appctl -t {} ovsdb-server/sync-status'.format(ctl_file)
            output = sh.execute(cmd, ssh_client=node.ssh_client, sudo=True)
            db_status = output.stdout
            if 'state: active' in db_status:
                status = 'active'
            elif 'state: backup' in db_status:
                status = 'backup'
            else:
                status = 'unknown'
            db_sync_status.setdefault(db, [])
            db_sync_status[db].append([node.hostname, status])
    LOG.debug('OVN DB status for all controllers: {}'.format(db_sync_status))
    return db_sync_status


def test_ovn_dbs_validations():
    if not neutron.has_ovn():
        LOG.debug('OVN not configured. OVN DB sync validations skipped')
        return

    test_case = tobiko.get_test_case()

    ovn_dbs_are_synchronized(test_case)
    ovn_dbs_vip_bindings(test_case)


class RAFTStatusError(tobiko.TobikoException):
    pass


def get_raft_cluster_details(hostname, database):
    """Return RAFT cluster details from specific node as dictionary"""
    if database not in OVNDBS:
        raise ValueError('{} database is not in the list {}'.format(
            database, OVNDBS))
    ctl_files = find_ovn_db_ctl_files()
    cmd = 'ovs-appctl -t {} cluster/status {}'.format(ctl_files[database],
                                                      DBNAMES[database])
    node_ssh = topology.get_openstack_node(hostname=hostname).ssh_client
    output = sh.execute(cmd, ssh_client=node_ssh, sudo=True).stdout
    cluster_status = {}
    for line in output.splitlines():
        if not re.match(r'^\s+', line):
            line_data = line.strip().split(':', 1)
            if len(line_data) == 1:
                continue
            section, data = line_data
            if not data.strip():
                cluster_status[section] = []
                continue
            cluster_status[section] = data.strip()
        else:
            cluster_status[section].append(line.strip())
    return cluster_status


def collect_raft_cluster_details(database):
    """Collect RAFT cluster details from all controllers

    Data collection should be restarted if one of the nodes has `candidate`
    role as it means that there is leader election in progress
    """
    for _ in tobiko.retry(timeout=30, interval=1):
        restart_collection = False
        cluster_details = []
        for node in topology.list_openstack_nodes(group='controller'):
            details = get_raft_cluster_details(node.hostname, database)
            if details['Role'].lower() == 'candidate':
                LOG.warning('Cluster not stable. Leader election in progress')
                LOG.debug('Cluster details from {} node:\n{}'.format(
                    node.hostname, details))
                restart_collection = True
                break
            cluster_details.append(details)
        if not restart_collection:
            break
    return cluster_details


def check_raft_timers(node_details):
    election_timer = int(node_details['Election timer'])
    leader_id = node_details['Leader']
    for srv_str in node_details['Servers']:
        if 'self' in srv_str:
            continue
        if node_details['Role'] == 'follower' and \
                not srv_str.startswith(leader_id):
            # There is no active connection between followers
            continue
        timer = re.findall(r'last msg [0-9]+ ms ago', srv_str)
        if len(timer) != 1:
            msg = 'Failed to parse connection timer from "{}"'.format(srv_str)
            LOG.error(msg)
            LOG.debug(node_details)
            raise RAFTStatusError(message=msg)
        if election_timer < int(timer[0].split()[2]):
            msg = 'Cluster communication time {} is higher then election'\
                  ' timer {}'.format(int(timer.split()[2]), election_timer)
            LOG.error(msg)
            LOG.debug(node_details)
            raise RAFTStatusError(message=msg)


def get_raft_ports():
    node = topology.list_openstack_nodes(group='controller')[0]
    ports = {}
    for db in OVNDBS:
        cluster_details = get_raft_cluster_details(node.hostname, db)
        _, port = parse_ips_from_db_connections(cluster_details['Address'])
        ports[db] = port
    return ports


def test_raft_cluster():
    if not _overcloud.is_ovn_using_raft():
        return
    test_case = tobiko.get_test_case()
    cluster_ports = get_raft_ports()
    for db in OVNDBS:
        cluster_details = collect_raft_cluster_details(db)
        leader_found = False
        for node_details in cluster_details:
            check_raft_timers(node_details)
            if node_details['Role'] == 'leader':
                test_case.assertFalse(leader_found)
                leader_found = True
        test_case.assertTrue(leader_found)
        for node in topology.list_openstack_nodes(group='controller'):
            node_ips = ip.list_ip_addresses(ssh_client=node.ssh_client)
            socs = ss.tcp_listening(port=cluster_ports[db],
                                    ssh_client=node.ssh_client)
            test_case.assertEqual(1, len(socs))
            test_case.assertIn(socs[0]['local_addr'], node_ips)
            test_case.assertEqual(socs[0]['process'][0], 'ovsdb-server')


def test_raft_clients_connected():
    """Verifies that all SBDB readers are connected to active nodes

    Unlike HA environment all operations are allowed to be performed to any
    available node. To have the better performance all heavy write operations
    are done to leader node, but readers are spreaded across all cluster
    controllers
    """
    test_case = tobiko.get_test_case()
    test_case.assertTrue(_overcloud.is_ovn_using_raft())
    db_con_str = get_ovn_db_connections()['sb']
    addrs, port = parse_ips_from_db_connections(db_con_str)
    for node_details in collect_raft_cluster_details('sb'):
        if node_details['Role'] == 'leader':
            leader_ips, _ = parse_ips_from_db_connections(
                    node_details['Address'])
            break
    test_case.assertIsNotNone(locals().get('leader_ips'))
    leader_ip = leader_ips[0]

    for node in topology.list_openstack_nodes(group='compute'):
        socs = ss.tcp_connected(dst_port=port, ssh_client=node.ssh_client)
        ovn_controller_found = False
        for soc in socs:
            if soc['process'][0] == 'ovn-controller':
                ovn_controller_found = True
            test_case.assertIn(soc['remote_addr'], addrs)
        test_case.assertTrue(ovn_controller_found)

    for node in topology.list_openstack_nodes(group='controller'):
        socs = ss.tcp_connected(dst_port=port, ssh_client=node.ssh_client)
        ref_processes = {'ovn-controller', 'neutron-server:', 'ovn-northd'}
        processes = set()
        for soc in socs:
            processes.add(soc['process'][0])
            if soc['process'][0] == 'ovn-northd':
                test_case.assertEqual(soc['remote_addr'], leader_ip)
            else:
                test_case.assertIn(soc['remote_addr'], addrs)
        test_case.assertEqual(processes, ref_processes)


def test_ovs_bridges_mac_table_size():
    test_case = tobiko.get_test_case()
    expected_mac_table_size = '50000'
    get_mac_table_size_cmd = ('ovs-vsctl get bridge {br_name} '
                              'other-config:mac-table-size')
    if neutron.has_ovn():
        get_br_mappings_cmd = ('ovs-vsctl get Open_vSwitch . '
                               'external_ids:ovn-bridge-mappings')
    else:
        get_br_mappings_cmd = (
            'crudini --get /var/lib/config-data/puppet-generated/neutron/'
            'etc/neutron/plugins/ml2/openvswitch_agent.ini '
            'ovs bridge_mappings')
    for node in topology.list_openstack_nodes(group='overcloud'):
        try:
            br_mappings_str = sh.execute(get_br_mappings_cmd,
                                         ssh_client=node.ssh_client,
                                         sudo=True).stdout.splitlines()[0]
        except sh.ShellCommandFailed:
            LOG.debug(f"bridge mappings not configured on node '{node.name}'",
                      exc_info=1)
            continue
        br_list = [br_mapping.split(':')[1] for br_mapping in
                   br_mappings_str.replace('"', '').split(',')]
        for br_name in br_list:
            mac_table_size = sh.execute(
                get_mac_table_size_cmd.format(br_name=br_name),
                ssh_client=node.ssh_client, sudo=True).stdout.splitlines()[0]
            test_case.assertEqual(mac_table_size.replace('"', ''),
                                  expected_mac_table_size)


OPENSTACK_NODE_GROUP = re.compile(r'(compute|controller|overcloud)')
OVS_NAMESPACE = re.compile(r'(qrouter.*|qdhcp.*|snat.*|fip.*)')


def test_ovs_namespaces_are_absent(
        group: typing.Pattern[str] = OPENSTACK_NODE_GROUP,
        namespace: typing.Pattern[str] = OVS_NAMESPACE):
    nodes = topology.list_openstack_nodes(group=group)

    namespaces: typing.Dict[str, typing.List[str]] = (
        collections.defaultdict(list))
    for node in nodes:
        for node_namespace in ip.list_network_namespaces(
                ssh_client=node.ssh_client, sudo=True):
            if namespace.match(node_namespace):
                namespaces[node.name].append(node_namespace)
    namespaces = dict(namespaces)

    test_case = tobiko.get_test_case()
    test_case.assertEqual(
        {}, dict(namespaces),
        f"OVS namespace(s) found on OpenStack nodes: {namespaces}")


OVS_INTERFACE = re.compile(r'(qvo.*|qvb.*|qbr.*|qr.*|qg.*|fg.*|sg.*)')


def test_ovs_interfaces_are_absent(
        group: typing.Pattern[str] = OPENSTACK_NODE_GROUP,
        interface: typing.Pattern[str] = OVS_INTERFACE):
    nodes = topology.list_openstack_nodes(group=group)

    interfaces: typing.Dict[str, typing.List[str]] = (
        collections.defaultdict(list))
    for node in nodes:
        for node_interface in ip.list_network_interfaces(
                ssh_client=node.ssh_client, sudo=True):
            if interface.match(node_interface):
                interfaces[node.name].append(node_interface)
    interfaces = dict(interfaces)

    test_case = tobiko.get_test_case()
    test_case.assertEqual(
        {}, interfaces,
        f"OVS interface(s) found on OpenStack nodes: {interfaces}")



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import functools
import json
import os
import typing

from oslo_log import log

import tobiko
from tobiko.openstack.keystone import _credentials
from tobiko.shell import find
from tobiko.shell import sh


LOG = log.getLogger(__name__)

YAML_SUFFIXES = ('.yaml', '.yml')
JSON_SUFFIXES = ('.json',)
CLOUDS_FILE_SUFFIXES = JSON_SUFFIXES + YAML_SUFFIXES


class CloudsFileNotFoundError(tobiko.TobikoException):
    message = "No such clouds file(s): {clouds_files!s}"


CloudsFileContentType = typing.Mapping[str, typing.Any]


class CloudsFileKeystoneCredentialsFixture(
        _credentials.KeystoneCredentialsFixture):

    def __init__(self,
                 credentials: _credentials.KeystoneCredentials = None,
                 connection: sh.ShellConnectionType = None,
                 environ: typing.Dict[str, str] = None,
                 cloud_name: str = None,
                 directories: typing.Iterable[str] = None,
                 filenames: typing.Iterable[str] = None):
        super().__init__(credentials=credentials,
                         connection=connection,
                         environ=environ)
        self._cloud_name = cloud_name
        if directories is not None:
            directories = list(directories)
        self._directories = directories
        if filenames is not None:
            filenames = list(filenames)
        self._filenames = filenames

    default_cloud_name: typing.Optional[str] = None

    @property
    def cloud_name(self) -> typing.Optional[str]:
        if self._cloud_name is None:
            self._cloud_name = self._get_cloud_name()
        return self._cloud_name

    @property
    def directories(self) -> typing.List[str]:
        if self._directories is None:
            directories = [self.connection.get_config_path(directory)
                           for directory in self._get_directories()]
            self._directories = directories
        return self._directories

    @property
    def filenames(self) -> typing.List[str]:
        if self._filenames is None:
            self._filenames = self._get_filenames()
        return self._filenames

    def _get_credentials(self) -> _credentials.KeystoneCredentials:
        try:
            filenames = find.find_files(path=self.directories,
                                        name=self.filenames,
                                        max_depth=1,
                                        type='f',
                                        ssh_client=self.connection.ssh_client)
        except find.FilesNotFound as ex:
            raise _credentials.NoSuchKeystoneCredentials(
                reason=('Cloud files not found:\n'
                        f"  login: {self.login}\n"
                        f"  directories: {self.directories}\n"
                        f"  filenames: {self.filenames}\n"
                        f"  error: {ex}\n")) from ex

        if self.cloud_name is None:
            raise _credentials.NoSuchKeystoneCredentials(
                reason=(f"[{self.fixture_name}] Clouds name not found at"
                        f" {self.login!r}"))

        for filename in filenames:
            file_spec = f"{self.login}:{filename}"
            content = load_clouds_file_content(
                connection=self.connection,
                filename=filename)
            try:
                return parse_credentials(
                    file_spec=file_spec,
                    content=content,
                    cloud_name=self.cloud_name)
            except _credentials.NoSuchKeystoneCredentials:
                LOG.debug(f'Cloud with name {self.cloud_name} not found '
                          f'in {file_spec}')
        raise _credentials.NoSuchKeystoneCredentials(
            reason=(f"[{self.fixture_name}] Keystone credentials not found "
                    f"for cloud name {self.cloud_name!r} in files "
                    f"{filenames!r} (login={self.login})"))

    def _get_cloud_name(self) -> typing.Optional[str]:
        for var_name in ['OS_CLOUD', 'OS_CLOUDNAME']:
            cloud_name = self.environ.get(var_name)
            if cloud_name:
                LOG.debug(f"Got cloud name from '{var_name}' environment "
                          f"variable: {cloud_name}", )
                return cloud_name
        return self._get_default_cloud_name()

    @staticmethod
    def _get_default_cloud_name() -> typing.Optional[str]:
        return tobiko.tobiko_config().keystone.cloud_name

    @staticmethod
    def _get_directories() -> typing.List[str]:
        return tobiko.tobiko_config().keystone.clouds_file_dirs

    @staticmethod
    def _get_filenames() -> typing.List[str]:
        return tobiko.tobiko_config().keystone.clouds_file_names


def parse_credentials(file_spec: str,
                      cloud_name: str,
                      content: CloudsFileContentType):
    clouds_section = content.get("clouds")
    if clouds_section is None:
        raise _credentials.NoSuchKeystoneCredentials(
            reason=f"'clouds' section not found in {file_spec!r}")

    clouds_config = clouds_section.get(cloud_name)
    if clouds_config is None:
        raise _credentials.NoSuchKeystoneCredentials(
            reason=f"cloud name {cloud_name!r} not found in {file_spec!r}")

    auth = clouds_config.get("auth")
    if auth is None:
        raise _credentials.NoSuchKeystoneCredentials(
            reason=f"'auth' section not found for cloud name "
                   f"{cloud_name!r} in {file_spec!r}")

    auth_url = auth.get("auth_url")
    if not auth_url:
        raise _credentials.NoSuchKeystoneCredentials(
            reason=f"'auth_url' is {auth_url!r} for cloud name "
                   f"{cloud_name!r} in {file_spec!r}")

    username = auth.get('username') or auth.get('user_id')
    if not username:
        raise _credentials.NoSuchKeystoneCredentials(
            reason=f"'username' is {username!r} for cloud name "
                   f"{cloud_name!r} in {file_spec!r}")

    password = auth.get('password')
    if not password:
        raise _credentials.NoSuchKeystoneCredentials(
            reason=f"'password' is {password!r} for cloud name "
                   f"{cloud_name!r} in {file_spec!r}")

    cacert = clouds_config.get('cacert')
    project_name = (auth.get('project_name') or
                    auth.get('tenant_namer') or
                    auth.get('project_id') or
                    auth.get_env('tenant_id'))

    api_version = (int(clouds_config.get("identity_api_version", 0)) or
                   _credentials.api_version_from_url(auth_url))
    if api_version == 2:
        return _credentials.keystone_credentials(
            api_version=api_version,
            auth_url=auth_url,
            username=username,
            password=password,
            project_name=project_name)
    else:
        domain_name = (auth.get("domain_name") or
                       auth.get("domain_id"))
        user_domain_name = (auth.get("user_domain_name") or
                            auth.get("user_domain_id"))
        project_domain_name = auth.get("project_domain_name")
        project_domain_id = auth.get("project_domain_id")
        trust_id = auth.get("trust_id")
        return _credentials.keystone_credentials(
            api_version=api_version,
            auth_url=auth_url,
            username=username,
            password=password,
            project_name=project_name,
            domain_name=domain_name,
            user_domain_name=user_domain_name,
            project_domain_name=project_domain_name,
            project_domain_id=project_domain_id,
            cacert=cacert,
            trust_id=trust_id)


@functools.lru_cache()
def load_clouds_file_content(connection: sh.ShellConnection,
                             filename: str) \
        -> CloudsFileContentType:
    with connection.open_file(filename, 'r') as f:
        _, suffix = os.path.splitext(filename)
        if suffix in JSON_SUFFIXES:
            LOG.debug(f'Load JSON clouds file: {filename!r}')
            return json.load(f)
        else:
            LOG.debug(f'Load YAML clouds file: {filename!r}')
            return tobiko.load_yaml(f)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import io
import typing

from paramiko import sftp_file

from tobiko import config
from tobiko.openstack import glance
from tobiko.openstack import neutron
from tobiko.openstack.stacks import _nova
from tobiko.shell import sh
from tobiko.shell import ssh
import tobiko.tripleo


CONF = config.CONF

CIRROS_IMAGE_VERSION = '0.5.2'

if tobiko.tripleo.has_undercloud():
    CIRROS_IMAGE_URL = (
        f"http://rhos-qe-mirror-rdu2.usersys.redhat.com/images/cirros-"
        f"{CIRROS_IMAGE_VERSION}-x86_64-disk.img")

else:
    CIRROS_IMAGE_URL = (
        'http://download.cirros-cloud.net/{version}/'
        'cirros-{version}-x86_64-disk.img').format(
        version=CIRROS_IMAGE_VERSION)


class CirrosImageFixture(glance.URLGlanceImageFixture):

    image_url = CONF.tobiko.cirros.image_url or CIRROS_IMAGE_URL
    image_name = CONF.tobiko.cirros.image_name
    image_file = CONF.tobiko.cirros.image_file
    container_format = CONF.tobiko.cirros.container_format or "bare"
    disk_format = CONF.tobiko.cirros.disk_format or "qcow2"
    username = CONF.tobiko.cirros.username or 'cirros'
    password = CONF.tobiko.cirros.password or 'gocubsgo'
    connection_timeout = CONF.tobiko.cirros.connection_timeout or 200.
    disabled_algorithms = CONF.tobiko.cirros.disabled_algorithms or {
        #: disabled_algorithms is required to connect to CirrOS servers
        # when using recent Paramiko versions (>= 2.9.2)
        'pubkeys': ['rsa-sha2-256', 'rsa-sha2-512']}


class CirrosFlavorStackFixture(_nova.FlavorStackFixture):
    ram = 128


class CirrosServerStackFixture(_nova.ServerStackFixture):

    #: Glance image used to create a Nova server instance
    image_fixture = tobiko.required_fixture(CirrosImageFixture)

    #: Flavor used to create a Nova server instance
    flavor_stack = tobiko.required_fixture(CirrosFlavorStackFixture)

    #: CirrOS can't get IP addresses from config-drive
    need_dhcp = True

    #: We expect CirrOS based servers to be fast to boot
    is_reachable_timeout: tobiko.Seconds = 300.

    @property
    def ssh_client(self) -> ssh.SSHClientFixture:
        ssh_client = super().ssh_client
        connection = sh.shell_connection(ssh_client)
        if not connection.is_cirros:
            connection = CirrosShellConnection(ssh_client=ssh_client)
            sh.register_shell_connection(connection)
        return ssh_client


class CirrosServerWithDefaultSecurityGroupStackFixture(
        CirrosServerStackFixture):

    #: Use "Default" Security Group of the project
    security_groups = [neutron.DEFAULT_SG_NAME]


class CirrosShellConnection(sh.SSHShellConnection):

    @property
    def is_cirros(self) -> bool:
        return True

    def get_file(self, remote_file: str, local_file: str):
        content = self.execute(f"cat '{remote_file}'").stdout
        with io.open(local_file, 'wt') as fd:
            fd.write(content)

    def put_file(self, local_file: str, remote_file: str):
        with io.open(local_file, 'rt') as fd:
            content = fd.read()
        self.execute(f"cat >'{remote_file}'", stdin=content)

    def open_file(self,
                  filename: typing.Union[str, bytes],
                  mode: str,
                  buffering: int = None) -> sftp_file.SFTPFile:
        raise NotImplementedError


class CirrosPeerServerStackFixture(CirrosServerStackFixture,
                                   _nova.PeerServerStackFixture):
    #: Peer server used to reach this one
    peer_stack = tobiko.required_fixture(CirrosServerStackFixture)


class CirrosSameHostServerStackFixture(
        CirrosPeerServerStackFixture, _nova.SameHostServerStackFixture):
    pass


class CirrosDifferentHostServerStackFixture(
        CirrosPeerServerStackFixture, _nova.DifferentHostServerStackFixture):
    pass


class RebootCirrosServerOperation(sh.RebootHostOperation):

    stack = tobiko.required_fixture(CirrosServerStackFixture)

    @property
    def ssh_client(self):
        return self.stack.ssh_client


class EvacuableCirrosImageFixture(CirrosImageFixture):

    tags = ['evacuable']


class EvacuableServerStackFixture(CirrosServerStackFixture):

    #: Glance image used to create a Nova server instance
    image_fixture = tobiko.required_fixture(EvacuableCirrosImageFixture)


class ExtraDhcpOptsCirrosServerStackFixture(CirrosServerStackFixture):
    use_extra_dhcp_opts = True
